# -*- coding: utf-8 -*-
# Copyright 2018-2024 the orix developers
#
# This file is part of orix.
#
# orix is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# orix is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with orix.  If not, see <http://www.gnu.org/licenses/>.

"""Generation of spherical grids in *S2*."""

from functools import partial
from typing import Callable, List, Mapping, Optional, Tuple

import numpy as np

from orix.sampling._polyhedral_sampling import (
    _compose_from_faces,
    _edge_grid_normalized_cube,
    _edge_grid_spherified_corner_cube,
    _edge_grid_spherified_edge_cube,
    _sample_length_equidistant,
)
from orix.vector import Vector3d


def _remove_pole_duplicates(
    azimuth: np.ndarray, polar: np.ndarray
) -> Tuple[np.ndarray, np.ndarray]:
    """Remove duplicate directions at the North (polar = 0) and South
    (polar = pi) poles from the grid on S2. In each case the direction
    with azimuth = 0 is kept.

    Parameters
    ----------
    azimuth
        Azimuth angles.
    polar
        Polar angles.

    Returns
    -------
    azimuth
        Azimuth angles without duplicates.
    polar
        Polar angles without duplicates.
    """
    mask_azimuth = azimuth > 0
    mask_polar_0 = np.isclose(polar, 0) * mask_azimuth
    mask_polar_pi = np.isclose(polar, np.pi) * mask_azimuth
    # create mask of vectors to keep
    mask = ~np.logical_or(mask_polar_0, mask_polar_pi)
    return azimuth[mask], polar[mask]


def _sample_S2_uv_mesh_coordinates(
    resolution: float,
    hemisphere: str = "both",
    offset: float = 0,
    azimuth_endpoint: bool = False,
) -> Tuple[np.ndarray, np.ndarray]:
    """Get spherical coordinates for UV mesh points on unit sphere *S2*.

    For more information see the docstring for
    :meth:`orix.sampling.S2_sampling.sample_S2_uv_mesh`.

    Parameters
    ----------
    resolution
        Maximum angle between nearest neighbour grid points, in degrees.
        The resolution of :math:`u` and :math:`v` are rounded up to get
        an integer number of equispaced polar and azimuthal grid lines.
    hemisphere
        Generate mesh points on either the ``"upper"``, ``"lower"`` or
        ``"both"`` hemispheres. Default is ``"both"``.
    offset
        Mesh points are offset in angular space by this fraction of the
        step size, must be in the range [0..1]. Default is 0.
    azimuth_endpoint
        If ``True`` then endpoint of the azimuth array is included in
        the calculation. Default is ``False``.

    Returns
    -------
    azimuth
        Azimuth angles.
    polar
        Polar angles.
    """
    hemisphere = hemisphere.lower()
    if hemisphere not in ("upper", "lower", "both"):
        raise ValueError('Hemisphere must be one of "upper", "lower", or "both".')

    if not 0 <= offset < 1:
        raise ValueError(
            "Offset is a fractional value of the angular step size "
            + "and must be in the range [0..1]."
        )

    if hemisphere == "both":
        polar_min = 0
        polar_max = 180
    elif hemisphere == "upper":
        polar_min = 0
        polar_max = 90
    elif hemisphere == "lower":
        polar_min = 90
        polar_max = 180
    polar_range = polar_max - polar_min
    # calculate steps in degrees to avoid rounding errors
    steps_azimuth = int(np.ceil(360 / resolution))
    steps_polar = int(np.ceil(polar_range / resolution)) + 1
    resolution = np.deg2rad(resolution)
    # calculate number of steps and step size angular spacing
    step_size_azimuth = (2 * np.pi) / steps_azimuth
    step_size_polar = np.deg2rad(polar_range) / (steps_polar - 1)

    azimuth = np.linspace(
        offset * step_size_azimuth,
        2 * np.pi + offset * step_size_azimuth,
        num=steps_azimuth,
        endpoint=azimuth_endpoint,
    )
    # convert to radians
    polar_min, polar_max = np.deg2rad(polar_min), np.deg2rad(polar_max)
    polar = np.linspace(
        polar_min + offset * step_size_polar,
        polar_max + offset * step_size_polar,
        num=steps_polar,
    )
    # polar coordinate cannot exceed polar_max
    polar = polar[polar <= polar_max]

    return azimuth, polar


def sample_S2_uv_mesh(
    resolution: float,
    hemisphere: str = "both",
    offset: float = 0,
    remove_pole_duplicates: bool = True,
) -> Vector3d:
    r"""Return vectors of a UV mesh on a unit sphere *S2*
    :cite:`cajaravelli2015four`.

    The mesh vertices are defined by the parametrization

    .. math::
        x &= \sin(u)\cos(v), \\
        y &= \sin(u)\sin(v), \\
        z &= \cos(u).

    Parameters
    ----------
    resolution
        Maximum angle between nearest neighbour grid points, in degrees.
        The resolution of :math:`u` and :math:`v` are rounded up to get
        an integer number of equispaced polar and azimuthal grid lines.
    hemisphere
        Generate mesh points on the ``"upper"``, ``"lower"`` or
        ``"both"`` hemispheres. Default is ``"both"``.
    offset
        Mesh points are offset in angular space by this fraction of the
        step size, must be in the range [0..1]. Default is 0.
    remove_pole_duplicates
        If ``True`` the duplicate mesh grid points at the North and
        South pole of the unit sphere are removed. Default is ``True``.

    Returns
    -------
    vec
        Vectors that sample the unit sphere.
    """
    azimuth, polar = _sample_S2_uv_mesh_coordinates(resolution, hemisphere, offset)
    azimuth_prod, polar_prod = np.meshgrid(azimuth, polar)

    if remove_pole_duplicates:
        azimuth_prod, polar_prod = _remove_pole_duplicates(azimuth_prod, polar_prod)

    return Vector3d.from_polar(azimuth=azimuth_prod, polar=polar_prod).unit


def _sample_S2_equal_area_coordinates(
    resolution: float,
    hemisphere: str = "both",
    azimuth_endpoint: bool = False,
    azimuth_range: Optional[Tuple[float, float]] = None,
    polar_range: Optional[Tuple[float, float]] = None,
) -> Tuple[np.ndarray, np.ndarray]:
    """Get spherical coordinates for equal area mesh points on unit
    sphere *S2*.

    For more information see the docstring for
    :meth:`orix.sampling.S2_sampling.sample_S2_equal_area_mesh`.

    Parameters
    ----------
    resolution
        The angular resolution in degrees of the azimuthal vectors.
    hemisphere
        Generate mesh points on the ``"upper"``, ``"lower"`` or
        ``"both"`` hemispheres. Default is ``"both"``.
    azimuth_endpoint
        If ``True`` then endpoint of the azimuth array is included in
        the calculation. Default is ``False``.
    azimuth_range, polar_range
        The (min, max) angular range for the azimuthal and polar
        coordinates, respectively, in radians. If provided then the
        ``hemisphere`` argument is ignored. Default is ``None``.

    Returns
    -------
    azimuth
        Azimuth angles.
    polar
        Polar angles.
    """
    # calculate number of steps and step size angular spacing
    # this parameter D in :cite:`rohrer2004distribution`.
    steps = int(np.ceil(90 / resolution))

    if azimuth_range is not None:
        azimuth_min, azimuth_max = azimuth_range
        if azimuth_min >= azimuth_max:
            raise ValueError(
                "`azimuth_range` requires values (min, max) where min < max."
            )
    else:
        # use full range
        azimuth_min, azimuth_max = 0, 2 * np.pi

    # no wrap around
    azimuth_min = max(azimuth_min, 0)
    azimuth_max = min(azimuth_max, 2 * np.pi)

    azimuth_range = azimuth_max - azimuth_min
    # azimuth should have 4D steps over range [0..2pi]
    azimuth_num = int(np.ceil(azimuth_range / (np.pi / 2) * steps))

    # polar coordinate is parameterized in terms of cos(theta)
    if polar_range is not None:
        polar_min, polar_max = polar_range
        # no wrap around
        polar_min = max(polar_min, 0)
        polar_max = min(polar_max, np.pi)
        if polar_min >= polar_max:
            raise ValueError(
                "`polar_range` requires values (min, max) where min < max."
            )
        # convert to units of cos(theta) for equal area spacing
        polar_min, polar_max = np.cos((polar_min, polar_max))
    else:
        hemisphere = hemisphere.lower()
        if hemisphere not in ("upper", "lower", "both"):
            raise ValueError("`hemisphere` must be one of 'upper', 'lower', or 'both'.")
        # polar_min and polar_max in units of cos(theta)
        if hemisphere == "both":
            polar_min = 1
            polar_max = -1
        elif hemisphere == "upper":
            polar_min = 1
            polar_max = 0
        elif hemisphere == "lower":
            polar_min = 0
            polar_max = -1

    polar_range = polar_min - polar_max  # opposite as cos([0..pi]) -> [1..-1]
    # polar should have D steps over range [0..pi/2] rad, ie. [1..0]
    # extra point as polar endpoint is True
    polar_num = int(np.ceil(polar_range * steps)) + 1

    # extra data point to account for endpoint
    if azimuth_endpoint:
        azimuth_num += 1

    azimuth = np.linspace(
        azimuth_min, azimuth_max, num=azimuth_num, endpoint=azimuth_endpoint
    )
    polar = np.linspace(
        polar_min,
        polar_max,
        num=polar_num,
    )
    polar = np.arccos(polar)

    return azimuth, polar


def sample_S2_equal_area_mesh(
    resolution: float,
    hemisphere: str = "both",
    remove_pole_duplicates: bool = True,
) -> Vector3d:
    """Return vectors of a cube mesh on a unit sphere *S2* according to
    equal area spacing :cite:`rohrer2004distribution`.

    Parameters
    ----------
    resolution
        The angular resolution in degrees of the azimuthal vectors.
    hemisphere
        Generate mesh points on the ``"upper"``, ``"lower"`` or
        ``"both"`` hemispheres. Default is ``"both"``.
    remove_pole_duplicates
        If ``True`` the duplicate mesh grid points at the North and
        South pole of the unit sphere are removed. If ``True`` then the
        returned vector has ``ndim = 1``, whereas ``ndim = 2`` (grid) if
        ``False``. Default is ``True``.

    Returns
    -------
    vec
        Vectors that sample the unit sphere.
    """
    azimuth, polar = _sample_S2_equal_area_coordinates(resolution, hemisphere)
    azimuth_prod, polar_prod = np.meshgrid(azimuth, polar)

    if remove_pole_duplicates:
        azimuth_prod, polar_prod = _remove_pole_duplicates(azimuth_prod, polar_prod)

    return Vector3d.from_polar(azimuth=azimuth_prod, polar=polar_prod).unit


def sample_S2_cube_mesh(
    resolution: float, grid_type: str = "spherified_corner"
) -> Vector3d:
    """Return vectors of a cube mesh projected on a unit sphere *S2*
    :cite:`cajaravelli2015four`.

    Parameters
    ----------
    resolution
        Maximum angle between neighbour grid points, in degrees.
    grid_type
        Type of cube grid: ``"normalized"``, ``"spherified_edge"`` or
        ``"spherified_corner"``.

    Returns
    -------
    vec
        Vectors that sample the unit sphere.

    Notes
    -----
    Vectors are sampled by projecting a grid on a cube onto the unit
    sphere. The mesh on the cube can be generated in a number of ways. A
    regular square grid with equidistant points corresponds to the
    ``"normalized"`` option. ``"spherified_edge"`` corresponds to points
    such that the row of vectors from the [001] to [011] is equiangular.
    ``"spherified_corner"`` corresponds to the case where the row of
    vectors from [001] to [111] is equiangular.
    """
    grid_type = grid_type.lower()
    grid_mapping = {
        "normalized": _edge_grid_normalized_cube,
        "spherified_edge": _edge_grid_spherified_edge_cube,
        "spherified_corner": _edge_grid_spherified_corner_cube,
    }
    try:
        grid_on_edge = grid_mapping[grid_type](resolution)
    except KeyError:
        raise ValueError(
            f"The `grid_type` {grid_type} is not among the valid options "
            f"{list(grid_mapping.keys())}"
        )

    x, y = np.meshgrid(grid_on_edge, grid_on_edge)
    x = x.ravel()
    y = y.ravel()
    z = np.ones(x.shape[0])

    # Grid on all faces of the cube, avoiding overlap of points on edges
    bottom = np.vstack([-x, -y, -z]).T
    top = np.vstack([x, y, z]).T
    east = np.vstack([z, x, -y]).T
    west = np.vstack([-z, -x, y]).T
    south = np.vstack([x, -z, y]).T
    north = np.vstack([-x, z, -y]).T

    # Two corners are missing with this procedure
    m_c = np.array([[-1, 1, 1], [1, -1, -1]])

    return Vector3d(np.vstack((bottom, top, east, west, south, north, m_c))).unit


def sample_S2_hexagonal_mesh(resolution: float) -> Vector3d:
    """Return vectors of a hexagonal bipyramid mesh projected on a unit
    sphere *S2*.

    Parameters
    ----------
    resolution
        Maximum angle between neighbour grid points, in degrees.

    Returns
    -------
    vec
        Vectors that sample the unit sphere.
    """
    number_of_steps = int(np.ceil(2 / np.tan(np.deg2rad(resolution))))
    if number_of_steps % 2 == 1:
        # an even number of steps is required to get a point in the middle
        # of the hexagon edge
        number_of_steps += 1
    grid_1D = _sample_length_equidistant(
        number_of_steps,
        length=1.0,
        include_end=True,
        positive_and_negative=False,
    )

    # top and bottom face of the hexagon
    axis_to_corner_1 = grid_1D[1:]
    axis_to_corner_2 = grid_1D
    u, v = np.meshgrid(axis_to_corner_1, axis_to_corner_2)
    u = u.ravel()
    v = v.ravel()

    # from square to hex lattice
    hexagon_edge_length = 2 / np.sqrt(3)
    transform = np.array([[hexagon_edge_length, hexagon_edge_length / 2], [0, 1]])
    uv = np.stack([u, v])
    xy = np.dot(transform, uv)
    x, y = xy

    # raise to pyramidal plane
    z = -1 / hexagon_edge_length * x - 1 / 2 * y + 1
    tolerance = -1e-7
    include_points = z > tolerance
    points_one_face = np.stack([coordinate[include_points] for coordinate in [x, y, z]])

    # repeat 6 times by rotating 60 degrees
    def rotation(r: np.ndarray) -> np.ndarray:
        return np.array(
            [[np.cos(r), -np.sin(r), 0], [np.sin(r), np.cos(r), 0], [0, 0, 1]]
        )

    angle = np.deg2rad(60)
    top_faces = np.hstack(
        [np.dot(rotation(i * angle), points_one_face) for i in range(6)]
    )
    bottom_faces = top_faces.copy()
    bottom_faces[2] *= -1
    exclude_rim = bottom_faces[2] < tolerance
    bottom_faces = bottom_faces.T[exclude_rim].T
    north_pole = np.array([[0, 0, 1]]).T
    south_pole = np.array([[0, 0, -1]]).T
    all_points = np.hstack([top_faces, north_pole, bottom_faces, south_pole])

    return Vector3d(all_points.T).unit


def sample_S2_random_mesh(resolution: float, seed: Optional[int] = None) -> Vector3d:
    """Return vectors of a random mesh on *S2*.

    Parameters
    ----------
    resolution
        The expected mean angle between nearest neighbor grid points in
        degrees.
    seed
        Passed to :func:`numpy.random.default_rng`, defaults to None
        which will give a "new" random result each time.

    Returns
    -------
    vec
        Vectors that sample the unit sphere.

    References
    ----------
    https://mathworld.wolfram.com/SpherePointPicking.html
    """
    # convert resolution in degrees to number of points
    number = int(1 / (4 * np.pi) * (360 / resolution) ** 2)
    rng = np.random.default_rng(seed=seed)
    xyz = rng.normal(size=(number, 3))
    return Vector3d(xyz).unit


def sample_S2_icosahedral_mesh(resolution: float) -> Vector3d:
    """Return vectors of an icosahedral mesh on *S2* :cite:`meshzoo`.

    Parameters
    ----------
    resolution
        Maximum angle between neighbour grid points, in degrees.

    Returns
    -------
    vec
        Vectors that sample the unit sphere.
    """
    t = (1.0 + np.sqrt(5.0)) / 2.0
    corners = np.array(
        [
            [-1, +t, +0],
            [+1, +t, +0],
            [-1, -t, +0],
            [+1, -t, +0],
            #
            [+0, -1, +t],
            [+0, +1, +t],
            [+0, -1, -t],
            [+0, +1, -t],
            #
            [+t, +0, -1],
            [+t, +0, +1],
            [-t, +0, -1],
            [-t, +0, +1],
        ]
    )
    faces = [
        (0, 11, 5),
        (0, 5, 1),
        (0, 1, 7),
        (0, 7, 10),
        (0, 10, 11),
        (1, 5, 9),
        (5, 11, 4),
        (11, 10, 2),
        (10, 7, 6),
        (7, 1, 8),
        (3, 9, 4),
        (3, 4, 2),
        (3, 2, 6),
        (3, 6, 8),
        (3, 8, 9),
        (4, 9, 5),
        (2, 4, 11),
        (6, 2, 10),
        (8, 6, 7),
        (9, 8, 1),
    ]
    # icosahedron edge length
    a = np.linalg.norm(corners[0]) / np.sin(2 * np.pi / 5)
    # icosahedron inscribed sphere radius
    r_i = np.sqrt(3) / 12 * (3 + np.sqrt(5)) * a
    n = int(np.ceil(a / (r_i * np.tan(np.deg2rad(resolution)))))
    vertices = _compose_from_faces(corners, faces, n)
    return Vector3d(vertices).unit


_sampling_method_registry: Mapping[str, Callable] = {
    "uv": sample_S2_uv_mesh,
    "equal_area": sample_S2_equal_area_mesh,
    "normalized_cube": partial(sample_S2_cube_mesh, grid_type="normalized"),
    "spherified_cube_edge": partial(sample_S2_cube_mesh, grid_type="spherified_edge"),
    "spherified_cube_corner": partial(
        sample_S2_cube_mesh, grid_type="spherified_corner"
    ),
    "icosahedral": sample_S2_icosahedral_mesh,
    "hexagonal": sample_S2_hexagonal_mesh,
    "random": sample_S2_random_mesh,
}
sampling_methods: List[str] = []
_sampling_method_names = set()
for sampling_name, sampling_method in _sampling_method_registry.items():
    sampling_methods.append(sampling_name)
    _func = (
        sampling_method.func
        if isinstance(sampling_method, partial)
        else sampling_method
    )
    _sampling_method_names.add(f":func:`orix.sampling.{_func.__name__}`")

_s2_sampling_docstring = (
    """Return unit vectors that sample S2 with a specific angular
    resolution.

    Parameters
    ----------
    resolution
        Maximum angle between nearest neighbour grid points, in degrees.
    method
        Sphere meshing method. Options are: {}. The default is
        ``\"spherified_cube_edge\"``.
    **kwargs
        Keyword arguments passed to the sampling function. For details
        see the sampling functions listed below.

    Returns
    -------
    vec
        Vectors that sample the unit sphere.

    See Also
    --------
    {}
    """
).format(
    ", ".join(map(lambda x: f'``"{x}"``', sampling_methods)),
    "\n    ".join(_sampling_method_names),
)


def sample_S2(
    resolution: float, method: str = "spherified_cube_edge", **kwargs
) -> Vector3d:
    try:
        sampling_method = _sampling_method_registry[method]
    except KeyError:
        raise NotImplementedError(
            f"Method not implemented. Valid options: {sampling_methods}"
        )
    return sampling_method(resolution, **kwargs)


setattr(sample_S2, "__doc__", _s2_sampling_docstring)



"""
======================
Select and modify data
======================

This example shows how to select data in a :class:`~orix.crystal_map.CrystalMap`, obtain
a new map from parts of the data and modify data inplace.

There are five ways to select data:
    1. NumPy slice(s)
    2. By one or more indices
    3. Per phase(s)
    4. Points considered "indexed"/"not_indexed"
    5. Boolean indexing
"""

from typing import List, Union

import matplotlib.pyplot as plt
import numpy as np

from orix import plot
from orix.crystal_map import CrystalMap, Phase


def plot_id(
    xmaps: Union[CrystalMap, List[CrystalMap]], titles: Union[str, List[str]]
) -> None:
    """Convenience function to plot at most four crystal maps showing
    rows, columns and IDs of each map point.
    """
    if isinstance(xmaps, CrystalMap):
        xmaps = [xmaps]
        titles = [titles]
    n_xmaps = len(xmaps)
    if n_xmaps > 2:
        fig_rows, fig_cols = 2, 2
    else:
        fig_rows, fig_cols = 1, len(xmaps)

    fig = plt.figure()
    for i in range(n_xmaps):
        ax = fig.add_subplot(fig_rows, fig_cols, i + 1, projection="plot_map")
        ax.plot_map(xmaps[i], "id", scalebar=False)
        rows, cols = xmaps[i].row, xmaps[i].col
        ax.set_xticks(np.arange(np.max(cols) + 1))
        ax.set_xticklabels(np.arange(np.max(cols) + 1))
        ax.set_yticks(np.arange(np.max(rows) + 1))
        ax.set_yticklabels(np.arange(np.max(rows) + 1))
        ax.set_xlabel("Column")
        ax.set_ylabel("Row")
        ax.set_title(titles[i])
        for j, r, c in zip(xmaps[i].id, rows, cols):
            ax.text(c, r, j, va="center", ha="center", c="r")
    fig.tight_layout()


########################################################################################
# We start by creating a crystal map with five rows and ten columns with all points
# having one phase and an identity rotation, and plot the row and column coordinates as
# well as the map ID of each point into the originally created map

xmap = CrystalMap.empty(shape=(5, 10))
xmap.phases[0].name = "a"
print(xmap)

plot_id(xmap, "Initial map")

########################################################################################
# Slice the map (1) by selecting some rows and columns. We'll plot the IDs again and see
# that these do not update after slicing. We'll also select some values from the sliced
# map (2) by passing one or more indices

xmap2 = xmap[1:4, 5:9]  # First number inclusive, last number exclusive
plot_id(xmap2, "Map slice")

print(xmap2[0, 0].id)
print(xmap2[-1].id)  # Last row
print(xmap2[:, 1].id)

########################################################################################
# Select data based on phase(s) (3) after adding a new phase to the phase list and
# giving some points in the data the new phase ID by modifying the phase IDs inplace

xmap.phases.add(Phase("b"))

xmap[1, 1].phase_id = 1
xmap[1:4, 5:9].phase_id = 1
print(xmap)

plot_id([xmap["a"], xmap["b"], xmap["a", "b"]], ["a", "b", "a and b"])

########################################################################################
# Set some points to not indexed and select only the indexed data or the not indexed
# data (4)

xmap[3:, 1:4].phase_id = -1
print(xmap)

plot_id([xmap["indexed"], xmap["not_indexed"]], ["Indexed", "Not indexed"])

########################################################################################
# Select data satisfying one or more criteria using boolean arrays (5)

plot_id(
    [
        xmap[xmap.id > 10],
        xmap[(xmap.phase_id == 0) & np.mod(xmap.id, 2).astype(bool)],
        xmap[(xmap.phase_id == 1) | ~xmap.is_indexed],
    ],
    ["Id greater than 10", "a and odd ID", "b or not indexed"],
)

########################################################################################
# When obtaining a new map from part of another map, the new map is a shallow copy of
# the initial map. This means that changes to the new map also changes the initial map.
# When this is undesirable, we make a deep copy by calling
# :meth:`~orix.crystal_map.CrystalMap.deepcopy`.

xmap3_shallow = xmap["b"]
xmap3_deep = xmap["b"].deepcopy()

xmap3_shallow[1, 5].phase_id = -1
xmap3_deep[1, 6].phase_id = -1

plot_id(
    [xmap["indexed"], xmap3_shallow["indexed"], xmap3_deep["indexed"]],
    ["Initial, indexed", "b shallow copy, indexed", "b deep copy, indexed"],
)



# -*- coding: utf-8 -*-
# Copyright 2018-2024 the orix developers
#
# This file is part of orix.
#
# orix is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# orix is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with orix.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import annotations

from typing import Any, Tuple, Union

import dask.array as da
from dask.diagnostics import ProgressBar
import numpy as np
from scipy.special import hyp0f1

from orix.quaternion import Quaternion
from orix.vector import Vector3d

# Used to round values below 1e-16 to zero
_FLOAT_EPS = np.finfo(float).eps


class Rotation(Quaternion):
    r"""Rotations of coordinate systems, leaving objects in place.

    Rotations :math:`R` can be parametrized in numerous ways, but in
    orix are handled as unit quaternions. Rotations can act on vectors
    or other rotations. They are often most easily visualized as being
    a turn of a certain angle :math:`\omega` about a certain axis
    :math:`\hat{\mathbf{n}}`.

    .. image:: /_static/img/rotation.png
       :width: 200px
       :alt: Rotation of an object illustrated with an axis and rotation angle.
       :align: center

    This rotation class add a sense of proper or improper rotations to
    :class:`Quaternion`. An improper rotation in orix operates on
    vectors as a rotation by the unit quaternion followed by inversion.

    See the documentation of quaternions for the applied conventions.

    Examples
    --------
    Rotate vector vA defined in coordinate system A to vector vB defined
    in coordinate system B

    >>> import numpy as np
    >>> from orix.quaternion import Quaternion
    >>> from orix.vector import Vector3d
    >>> R = Rotation.random()
    >>> vA = Vector3d.random()
    >>> vB = R * vA
    >>> np.allclose(vB.data, np.dot(R.to_matrix().squeeze(), vA.data.squeeze()))
    True

    Combine two rotations R1 and R2, applied in that order

    >>> R1, R2 = Rotation.random(2)
    >>> R12 = R2 * R1
    >>> np.allclose(
    ...     R12.to_matrix().squeeze(),
    ...     np.dot(R2.to_matrix().squeeze(), R1.to_matrix().squeeze())
    ... )
    True
    """

    def __init__(self, data: Union[np.ndarray, Rotation, list, tuple]):
        super().__init__(data)
        self._data = np.concatenate((self.data, np.zeros(self.shape + (1,))), axis=-1)
        if isinstance(data, Rotation):
            self.improper = data.improper
        with np.errstate(divide="ignore", invalid="ignore"):
            self.data /= self.norm[..., np.newaxis]

    # -------------------------- Properties -------------------------- #

    @property
    def improper(self) -> np.ndarray:
        """Return ``True`` for improper rotations and ``False``
        otherwise.
        """
        return self._data[..., -1].astype(bool)

    @improper.setter
    def improper(self, value: np.ndarray):
        self._data[..., -1] = value

    @property
    def antipodal(self) -> Rotation:
        """Return the rotation and its antipodal."""
        R = self.__class__(np.stack([self.data, -self.data]))
        R.improper = self.improper
        return R

    # ------------------------ Dunder methods ------------------------ #

    def __mul__(
        self, other: Union[Rotation, Quaternion, Vector3d, np.ndarray, int, list]
    ):
        # Combine rotations self * other as first other, then self
        if isinstance(other, Rotation):
            Q = Quaternion(self) * Quaternion(other)
            R = other.__class__(Q)
            R.improper = np.logical_xor(self.improper, other.improper)
            return R
        if isinstance(other, Quaternion):
            Q = Quaternion(self) * other
            return Q
        if isinstance(other, Vector3d):
            v = Quaternion(self) * other
            improper = (self.improper * np.ones(other.shape)).astype(bool)
            v[improper] = -v[improper]
            return v
        if isinstance(other, int) or isinstance(other, list):  # abs(1)
            other = np.atleast_1d(other).astype(int)
            if isinstance(other, np.ndarray):
                if not np.all(abs(other) == 1):
                    raise ValueError("Rotations can only be multiplied by 1 or -1")
                R = Rotation(self.data)
                R.improper = np.logical_xor(self.improper, other == -1)
                return R
        return NotImplemented

    def __neg__(self) -> Rotation:
        R = self.__class__(self.data)
        R.improper = np.logical_not(self.improper)
        return R

    def __getitem__(self, key) -> Rotation:
        R = super().__getitem__(key)
        R.improper = self.improper[key]
        return R

    def __invert__(self) -> Rotation:
        R = super().__invert__()
        R.improper = self.improper
        return R

    def __eq__(self, other: Union[Any, Rotation]) -> bool:
        """Check if the rotations have equal shapes and values."""
        if (
            isinstance(other, Rotation)
            and self.shape == other.shape
            and np.allclose(self.data, other.data)
            and np.allclose(self.improper, other.improper)
        ):
            return True
        else:
            return False

    # ------------------------ Class methods ------------------------- #

    @classmethod
    def random_vonmises(
        cls,
        shape: Union[int, tuple] = (1,),
        alpha: float = 1.0,
        reference: Union[list, tuple, Rotation] = (1, 0, 0, 0),
    ) -> Rotation:
        """Return random rotations with a simplified Von Mises-Fisher
        distribution.

        Parameters
        ----------
        shape
            The shape of the required object.
        alpha
            Parameter for the VM-F distribution. Lower values lead to
            "looser" distributions.
        reference
            The center of the distribution.

        Returns
        -------
        R
            Rotations.
        """
        shape = (shape,) if isinstance(shape, int) else shape
        reference = Rotation(reference)
        n = int(np.prod(shape))
        sample_size = int(alpha) * n
        R = []
        f_max = von_mises(reference, alpha, reference)
        while len(R) < n:
            R_i = cls.random(sample_size)
            f = von_mises(R_i, alpha, reference)
            x = np.random.rand(sample_size)
            R_i = R_i[x * f_max < f]
            R += list(R_i)
        return cls.stack(R[:n]).reshape(shape)

    # --------------------- Other public methods --------------------- #

    def unique(
        self,
        return_index: bool = False,
        return_inverse: bool = False,
        antipodal: bool = True,
    ) -> Union[
        Rotation,
        Tuple[Rotation, np.ndarray],
        Tuple[Rotation, np.ndarray, np.ndarray],
    ]:
        """Return the unique rotations from these rotations.

        Two rotations are not unique if they have the same propriety
        AND:
        - they have the same numerical value OR
        - the numerical value of one is the negative of the other

        Parameters
        ----------
        return_index
            If ``True``, will also return the indices of the (flattened)
            data where the unique entries were found.
        return_inverse
            If ``True``, will also return the indices to reconstruct the
            (flattened) data from the unique data.
        antipodal
            If ``False``, rotations representing the same transformation
            whose values are numerically different (negative) will *not*
            be considered unique.

        Returns
        -------
        R
            Unique rotations.
        idx_sort
            Indices of the flattened rotations where the unique entries
            are found. Only returned if ``return_index=True``.
        inv
            Indices to reconstruct the flattened rotations from the
            initial rotations. Only returned if ``return_inverse=True``.
        """
        if self.size == 0:
            return self.empty()

        R = self.flatten()

        if antipodal:
            abcd = R._differentiators()
        else:
            abcd = np.stack([R.a, R.b, R.c, R.d, R.improper], axis=-1).round(10)
        _, idx, inv = np.unique(abcd, axis=0, return_index=True, return_inverse=True)
        idx_argsort = np.argsort(idx)
        idx_sort = idx[idx_argsort]

        # Build inverse index map
        inv_map = np.empty_like(idx_argsort)
        inv_map[idx_argsort] = np.arange(idx_argsort.size)
        inv = inv_map[inv]
        dat = R[idx_sort]
        dat.improper = R.improper[idx_sort]

        if return_index and return_inverse:
            return dat, idx_sort, inv
        elif return_index and not return_inverse:
            return dat, idx_sort
        elif return_inverse and not return_index:
            return dat, inv
        else:
            return dat

    def angle_with(self, other: Rotation, degrees: bool = False) -> np.ndarray:
        """Return the angles of rotation transforming the rotations to
        the other rotations.

        Parameters
        ----------
        other
            Other rotations.
        degrees
            If ``True``, the angles are returned in degrees. Default is
            ``False``.

        Returns
        -------
        angles
            Angles of rotation in radians (``degrees=False``) or degrees
            (``degrees=True``).

        See Also
        --------
        angle_with
        Orientation.angle_with
        """
        other = Rotation(other)
        dot_products = self.unit.dot(other.unit)
        # Round because some dot products are slightly above 1
        dot_products = np.round(dot_products, np.finfo(dot_products.dtype).precision)
        angles = np.nan_to_num(np.arccos(2 * dot_products**2 - 1))
        if degrees:
            angles = np.rad2deg(angles)
        return angles

    def angle_with_outer(self, other: Rotation, degrees: bool = False) -> np.ndarray:
        """Return the angles of rotation transforming the rotations to
        all the other rotations.

        Parameters
        ----------
        other
            Another rotation.
        degrees
            If ``True``, the angles are returned in degrees. Default is
            ``False``.

        Returns
        -------
        angles
            Angles of rotation in radians (``degrees=False``) or degrees
            (``degrees=True``).

        Examples
        --------
        >>> from orix.quaternion import Rotation
        >>> R1 = Rotation.random((5, 3))
        >>> R2 = Rotation.random((6, 2))
        >>> omega = R1.angle_with_outer(R2)
        >>> omega.shape
        (5, 3, 6, 2)

        See Also
        --------
        angle_with
        Orientation.angle_with_outer
        """
        dot_products = self.unit.dot_outer(other.unit)
        angles = np.nan_to_num(np.arccos(2 * dot_products**2 - 1))
        if degrees:
            angles = np.rad2deg(angles)
        return angles

    def outer(
        self,
        other: Union[Rotation, Vector3d],
        lazy: bool = False,
        chunk_size: int = 20,
        progressbar: bool = True,
    ) -> Union[Rotation, Vector3d]:
        """Return the outer rotation products of the rotations and the
        other rotations or vectors.

        Parameters
        ----------
        other
            Other rotations or vectors.
        lazy
            Whether to compute the outer products using :mod:`dask`.
            This option can be used to reduce memory usage when working
            with large arrays. Default is ``False``.
        chunk_size
            Number of rotations per axis to include in each iteration of
            the computation. Default is 20. Only applies when
            ``lazy=True``. Increasing this might reduce the computation
            time at the cost of increased memory use.
        progressbar
            Whether to show a progressbar during computation if
            ``lazy=True``. Default is ``True``.

        Returns
        -------
        R
            Outer rotation products.
        """
        if lazy:
            darr = self._outer_dask(other, chunk_size=chunk_size)
            arr = np.empty(darr.shape)
            if progressbar:
                with ProgressBar():
                    da.store(darr, arr)
            else:
                da.store(darr, arr)
            R = other.__class__(arr)
        else:
            R = super().outer(other)

        if isinstance(R, Rotation):
            R.improper = np.logical_xor.outer(self.improper, other.improper)
        elif isinstance(R, Vector3d):
            R[self.improper] = -R[self.improper]

        return R

    def flatten(self) -> Rotation:
        """Return a new rotation instance collapsed into one dimension.

        Returns
        -------
        R
            Rotations collapsed into one dimension.
        """
        R = super().flatten()
        R.improper = self.improper.T.flatten().T
        return R

    def dot_outer(self, other: Rotation) -> np.ndarray:
        """Return the outer dot products of the rotations and the other
        rotations.

        Parameters
        ----------
        other
            Other rotations.

        Returns
        -------
        cosines
            Outer dot products.
        """
        dot_products = np.abs(super().dot_outer(other))
        if isinstance(other, Rotation):
            improper = self.improper.reshape(self.shape + (1,) * len(other.shape))
            i = np.logical_xor(improper, other.improper)
            dot_products = np.minimum(~i, dot_products)
        else:
            dot_products[self.improper] = 0
        return dot_products

    def inv(self) -> Quaternion:
        r"""Return the inverse rotations :math:`R^{-1}`."""
        return self.__invert__()

    # -------------------- Other private methods --------------------- #

    def _differentiators(self) -> np.ndarray:
        a = self.a
        b = self.b
        c = self.c
        d = self.d
        i = self.improper
        # fmt: off
        abcd = np.stack(
            (
                a ** 2, b ** 2, c ** 2, d ** 2,
                a * b, a * c, a * d,
                b * c, b * d,
                c * d,
                i,
            ),
            axis=-1,
        ).round(12)
        # fmt: on
        return abcd


def von_mises(
    x: Rotation, alpha: float, reference: Rotation = Rotation((1, 0, 0, 0))
) -> np.ndarray:
    r"""A vastly simplified Von Mises-Fisher distribution calculation.

    Parameters
    ----------
    x
        Rotations.
    alpha
        Lower values of alpha lead to "looser" distributions.
    reference
        Reference rotation. Default is the identity rotation.

    Returns
    -------
    R
        Rotations.

    Notes
    -----
    This simplified version of the distribution is calculated using

    .. math::
        \frac{\exp\left(2\alpha\cos\left(\omega\right)\right)}{\_0F\_1\left(\frac{N}{2}, \alpha^2\right)}

    where :math:`\omega` is the angle between orientations and :math:`N`
    is the number of relevant dimensions, in this case 3.
    """
    angle = Rotation(x).angle_with(reference)
    return np.exp(2 * alpha * np.cos(angle.data)) / hyp0f1(1.5, alpha**2)



# -*- coding: utf-8 -*-
# Copyright 2018-2024 the orix developers
#
# This file is part of orix.
#
# orix is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# orix is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with orix.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import annotations

from copy import deepcopy
from itertools import product
from typing import Optional, Tuple, Union

from diffpy.structure import Lattice
import numpy as np

from orix.vector import Vector3d


class Miller(Vector3d):
    r"""Direct crystal lattice vectors (uvw or UVTW) and reciprocal
    crystal lattice vectors (hkl or hkil), the latter known as Miller
    indices, describing directions with respect to the crystal reference
    frame defined by a phase's crystal lattice and symmetry.

    Exactly one of ``xyz``, ``uvw``, ``UVTW``, ``hkl``, or ``hkil``
    must be passed.

    The vectors are stored internally as cartesian coordinates in
    :attr:`data`.

    Parameters
    ----------
    xyz
        Vector(s) given in cartesian coordinates. Default is ``None``.
    uvw
        Indices of direct lattice vector(s). Default is ``None``.
    UVTW
        Indices of direct lattice vector(s), often preferred over
        ``uvw`` in trigonal and hexagonal lattices. Default is ``None``.
    hkl
        Indices of reciprocal lattice vector(s). Default is ``None``.
    hkil
        Indices of reciprocal lattice vector(s), often preferred
        over ``hkl`` in trigonal and hexagonal lattices. Default is
        ``None``.
    phase
        A phase with a crystal lattice and symmetry. Must be passed
        whenever direct or reciprocal lattice vectors are created.

    Notes
    -----
    The Miller-Bravais indices :math:`UVTW` are defined as

    .. math::
        U &= \frac{2u - v}{3}, \\
        V &= \frac{2v - u}{3}, \\
        T &= -\frac{u + v}{3}, \\
        W &= w.
    """

    def __init__(
        self,
        xyz: Union[np.ndarray, list, tuple, None] = None,
        uvw: Union[np.ndarray, list, tuple, None] = None,
        UVTW: Union[np.ndarray, list, tuple, None] = None,
        hkl: Union[np.ndarray, list, tuple, None] = None,
        hkil: Union[np.ndarray, list, tuple, None] = None,
        phase: Optional["orix.crystal_map.Phase"] = None,
    ):
        """Create a set of direct lattice vectors (uvw or UVTW) or
        reciprocal lattice vectors (hkl or hkil) describing directions
        with respect to a crystal reference frame defined by a phase's
        crystal lattice and symmetry.
        """
        n_passed = np.sum([i is not None for i in [xyz, uvw, UVTW, hkl, hkil]])
        if n_passed == 0 or n_passed > 1:
            raise ValueError(
                "Exactly one of `xyz`, `uvw`, `UVTW`, `hkl`, `hkil` must be passed"
            )
        if xyz is None and phase is None:
            raise ValueError(
                "A phase with a crystal lattice and symmetry must be passed to create "
                "direct or reciprocal lattice vector(s)"
            )

        self.phase = phase
        if xyz is not None:
            xyz = np.asarray(xyz)
            self.coordinate_format = "xyz"
        elif uvw is not None:
            xyz = _transform_space(uvw, "d", "c", phase.structure.lattice)
            self.coordinate_format = "uvw"
        elif UVTW is not None:
            UVTW = np.asarray(UVTW)
            _check_UVTW(UVTW)
            uvw = _UVTW2uvw(UVTW=UVTW)
            xyz = _transform_space(uvw, "d", "c", phase.structure.lattice)
            self.coordinate_format = "UVTW"
        elif hkl is not None:
            xyz = _transform_space(hkl, "r", "c", phase.structure.lattice)
            self.coordinate_format = "hkl"
        elif hkil is not None:
            hkil = np.asarray(hkil)
            _check_hkil(hkil)
            hkl = _hkil2hkl(hkil)
            xyz = _transform_space(hkl, "r", "c", phase.structure.lattice)
            self.coordinate_format = "hkil"
        super().__init__(xyz)

    # -------------------------- Properties -------------------------- #

    @property
    def coordinate_format(self) -> str:
        """Return or set the vector coordinate format.

        Parameters
        ----------
        value : str
            Vector coordinate format, either ``"xyz"``, ``"uvw"``,
            ``"UVTW"``, ``"hkl"``, or ``"hkil"``.
        """
        return self._coordinate_format

    @coordinate_format.setter
    def coordinate_format(self, value: str):
        """Set the vector coordinate format."""
        formats = ["xyz", "uvw", "UVTW", "hkl", "hkil"]
        if value not in formats:
            raise ValueError(f"Available coordinate formats are {formats}")
        self._coordinate_format = value

    @property
    def coordinates(self) -> np.ndarray:
        """Return the vector coordinates."""
        coordinate_format = self.coordinate_format
        if coordinate_format == "xyz":
            coordinate_format = "data"
        coordinates = self.__getattribute__(coordinate_format)
        return coordinates

    @property
    def hkl(self) -> np.ndarray:
        """Return or set the reciprocal lattice vectors.

        Parameters
        ----------
        value : np.ndarray
            New reciprocal lattice vector array.
        """
        return _transform_space(self.data, "c", "r", self.phase.structure.lattice)

    @hkl.setter
    def hkl(self, value: np.ndarray):
        """Set the reciprocal lattice vectors."""
        self.data = _transform_space(value, "r", "c", self.phase.structure.lattice)

    @property
    def hkil(self):
        r"""Return or set the reciprocal lattice vectors expressed as
        4-index Miller-Bravais indices.

        Parameters
        ----------
        value : np.ndarray
            New reciprocal lattice vector array. The sum of the first
            three indices, :math:`h`, :math:`k`, and :math:`i` must be
            zero.
        """
        return _hkl2hkil(self.hkl)

    @hkil.setter
    def hkil(self, value: np.ndarray):
        """Set the reciprocal lattice vectors expressed as 4-index
        Miller-Bravais indices.
        """
        self.hkl = _hkil2hkl(value)

    @property
    def h(self) -> np.ndarray:
        """Return the first reciprocal lattice vector index."""
        return self.hkl[..., 0]

    @property
    def k(self) -> np.ndarray:
        """Return the second reciprocal lattice vector index."""
        return self.hkl[..., 1]

    @property
    def i(self) -> np.ndarray:
        r"""Return the third reciprocal lattice vector index in 4-index
        Miller-Bravais indices, equal to :math:`-(h + k)`.
        """
        return self.hkil[..., 2]

    @property
    def l(self) -> np.ndarray:
        """Return the third reciprocal lattice vector index, or fourth
        index in 4-index Miller Bravais indices.
        """
        return self.hkl[..., 2]

    @property
    def uvw(self) -> np.ndarray:
        """Return or set the direct lattice vectors.

        Parameters
        ----------
        value : np.ndarray
            New direct lattice vector array.
        """
        return _transform_space(self.data, "c", "d", self.phase.structure.lattice)

    @uvw.setter
    def uvw(self, value: np.ndarray):
        """Set the direct lattice vectors."""
        self.data = _transform_space(value, "d", "c", self.phase.structure.lattice)

    @property
    def UVTW(self):
        r"""Return or set the direct lattice vectors expressed as
        4-index Weber symbols.

        They are defined as

        .. math::
            U &= \frac{2u - v}{3}, \\
            V &= \frac{2v - u}{3}, \\
            T &= -\frac{u + v}{3}, \\
            W &= w.

        Parameters
        ----------
        value : np.ndarray
            New direct lattice vector array. The sum of the first three
            indices, :math:`U`, :math:`V`, and :math:`T` must be zero.
        """
        return _uvw2UVTW(self.uvw)

    @UVTW.setter
    def UVTW(self, value):
        """Set the direct lattice vectors expressed as 4-index Weber
        symbols.
        """
        self.uvw = _UVTW2uvw(value)

    @property
    def u(self) -> np.ndarray:
        """Return the first direct lattice vector index."""
        return self.uvw[..., 0]

    @property
    def v(self) -> np.ndarray:
        """Return the second direct lattice vector index."""
        return self.uvw[..., 1]

    @property
    def w(self) -> np.ndarray:
        """Return the third direct lattice vector index."""
        return self.uvw[..., 2]

    @property
    def U(self) -> np.ndarray:
        r"""Return the first direct lattice vector index in 4-index
        Weber symbols, equal to :math:`(2u - v)/(3)`.
        """
        return self.UVTW[..., 0]

    @property
    def V(self) -> np.ndarray:
        r"""Return the second direct lattice vector index in 4-index
        Weber symbols, equal to :math:`(2v - u)/3`.
        """
        return self.UVTW[..., 1]

    @property
    def T(self) -> np.ndarray:
        r"""Return the third direct lattice vector index in 4-index
        Weber symbols, equal to :math:`-(u + v)/3`.
        """
        return self.UVTW[..., 2]

    @property
    def W(self) -> np.ndarray:
        r"""Return the fourth direct lattice vector index in 4-index
        Weber symbols, equal to :math:`w`.
        """
        return self.UVTW[..., 3]

    @property
    def length(self) -> np.ndarray:
        """Return the length of each vector given in lattice parameter
        units if the :attr:`coordinate_format` attribute equals
        ``"uvw"`` or ``"UVTW"``, and inverse lattice parameter units if
        the attribute equals ``"hkl"`` or ``"hkil"``.

        If the attribute equals ``"xyz"``, the norms of the vectors in
        :attr:`data` are returned.
        """
        if self.coordinate_format in ["hkl", "hkil"]:
            return self.phase.structure.lattice.rnorm(self.hkl)
        elif self.coordinate_format in ["uvw", "UVTW"]:
            return self.phase.structure.lattice.norm(self.uvw)
        else:
            return self.norm

    @property
    def multiplicity(self) -> np.ndarray:
        """Return the number of symmetrically equivalent directions per
        vector.
        """
        _, l = self.symmetrise(unique=True, return_multiplicity=True)
        return l.reshape(self.shape)

    @property
    def space(self) -> str:
        """Return whether the vector is in direct (``"d"``) or
        reciprocal (``"r"``) space.
        """
        if self.coordinate_format in ["xyz", "uvw", "UVTW"]:
            return "d"
        else:
            return "r"

    @property
    def is_hexagonal(self) -> bool:
        """Return whether the crystal reference frame is
        hexagonal/trigonal.
        """
        return self.phase.is_hexagonal

    @property
    def unit(self) -> Miller:
        """Return unit vectors."""
        m = self.__class__(xyz=super().unit.data, phase=self.phase)
        m.coordinate_format = self.coordinate_format
        return m

    # ------------------------ Dunder methods ------------------------ #

    def __repr__(self) -> str:
        """String representation."""
        name = self.__class__.__name__
        shape = self.shape
        symmetry = None if self.phase is None else self.phase.point_group.name
        coordinate_format = self.coordinate_format
        data = np.array_str(self.coordinates, precision=4, suppress_small=True)
        return (
            f"{name} {shape}, point group {symmetry}, {coordinate_format}\n" f"{data}"
        )

    def __getitem__(self, key) -> Miller:
        """NumPy fancy indexing of vectors."""
        m = self.__class__(xyz=self.data[key], phase=self.phase).deepcopy()
        m.coordinate_format = self.coordinate_format
        return m

    # ------------------------ Class methods ------------------------- #

    @classmethod
    def from_highest_indices(
        cls,
        phase: "orix.crystal_map.Phase",
        uvw: Union[np.ndarray, list, tuple, None] = None,
        hkl: Union[np.ndarray, list, tuple, None] = None,
    ) -> Miller:
        """Create a set of unique direct or reciprocal lattice vectors
        from three highest indices and a phase (crystal lattice and
        symmetry).

        Parameters
        ----------
        phase
            A phase with a crystal lattice and symmetry.
        uvw
            Three highest direct lattice vector indices.
        hkl
            Three highest reciprocal lattice vector indices.
        """
        if uvw is not None:
            coordinate_format = "uvw"
            highest_idx = uvw
        elif hkl is not None:
            coordinate_format = "hkl"
            highest_idx = hkl
        else:
            raise ValueError("Either highest `hkl` or `uvw` indices must be passed")
        idx = _get_indices_from_highest(highest_indices=highest_idx)
        init_kw = {coordinate_format: idx, "phase": phase}
        return cls(**init_kw).unique()

    @classmethod
    def from_min_dspacing(
        cls, phase: "orix.crystal_map.Phase", min_dspacing: float = 0.05
    ) -> Miller:
        """Create a set of unique reciprocal lattice vectors with a
        a direct space interplanar spacing greater than a lower
        threshold.

        Parameters
        ----------
        phase
            A phase with a crystal lattice and symmetry.
        min_dspacing
            Smallest interplanar spacing to consider. Default is 0.05,
            in the unit used to define the lattice parameters in
            ``phase``.
        """
        highest_hkl = _get_highest_hkl(
            lattice=phase.structure.lattice, min_dspacing=min_dspacing
        )
        hkl = _get_indices_from_highest(highest_indices=highest_hkl)
        hkl = hkl.astype(float).round(0)
        return cls(hkl=hkl, phase=phase).unique()

    @classmethod
    def random(
        cls,
        phase: "orix.crystal_map.Phase",
        shape: Union[int, tuple] = 1,
        coordinate_format: str = "xyz",
    ) -> Miller:
        """Create random Miller indices.

        Parameters
        ----------
        phase
            A phase with a crystal lattice and symmetry.
        shape
            Shape of the indices.
        coordinate_format
            Coordinate format of indices, either ``"xyz"`` (default),
            ``"uvw"``, ``"UVTW"``, ``"hkl"``, or ``"hkil"``.

        Returns
        -------
        m
            Random Miller indices.

        Examples
        --------
        >>> from orix.crystal_map import Phase
        >>> from orix.vector import Miller
        >>> phase = Phase(point_group="m-3m")
        >>> _ = Miller.random(phase)
        >>> _ = Miller.random(phase, (3, 4))
        >>> _ = Miller.random(phase, (3, 4), "hkl")
        """
        v = Vector3d.random(shape)
        m = Miller(xyz=v.data, phase=phase)
        m.coordinate_format = coordinate_format
        return m

    # --------------------- Other public methods --------------------- #

    def deepcopy(self) -> Miller:
        """Return a deepcopy of the instance."""
        return deepcopy(self)

    def round(self, max_index: int = 20) -> Miller:
        """Round a set of index triplet (Miller) or quartet
        (Miller-Bravais/Weber) to the *closest* smallest integers.

        Adopted from MTEX' :code:`Miller.round` function.

        Parameters
        ----------
        max_index
            Maximum integer index to round to, by default 20.

        Returns
        -------
        mill
            Rounded set of index triplet(s) or quartet(s).
        """
        if self.coordinate_format == "xyz":
            return self.deepcopy()
        else:
            new_coords = _round_indices(indices=self.coordinates, max_index=max_index)
            init_kw = {self.coordinate_format: new_coords, "phase": self.phase}
            return self.__class__(**init_kw)

    def symmetrise(
        self,
        unique: bool = False,
        return_multiplicity: bool = False,
        return_index: bool = False,
    ) -> Union[
        Miller, Tuple[Miller, np.ndarray], Tuple[Miller, np.ndarray, np.ndarray]
    ]:
        """Return vectors symmetrically equivalent to the vectors.

        Parameters
        ----------
        unique
            Whether to return only unique vectors. Default is ``False``.
        return_multiplicity
            Whether to return the multiplicity of each vector. Default
            is ``False``.
        return_index
            Whether to return the index into the vectors for the
            returned symmetrically equivalent vectors. Default is
            ``False``.

        Returns
        -------
        m
            Flattened symmetrically equivalent vectors.
        multiplicity
            Multiplicity of each vector. Returned if
            ``return_multiplicity=True``.
        idx
            Index into the vectors for the returned symmetrically
            equivalent vectors. Returned if ``return_index=True``.
        """
        if return_multiplicity and not unique:
            raise ValueError("`unique` must be True when `return_multiplicity` is True")
        elif return_index and not unique:
            raise ValueError("`unique` must be True when `return_index` is True")

        # Symmetrise directions with respect to crystal symmetry on the
        # flattened set of vectors
        operations = self.phase.point_group
        v2 = operations.outer(self.flatten())

        if unique:
            n_v = self.size  # Number of initial vectors in `self`

            # Array for symmetrically equivalent vectors
            v3 = self.zero((n_v, operations.size))

            # Array for multiplicity of initial vectors
            multiplicity = np.zeros(n_v, dtype=int)

            # Array for index into `self` for the returned symmetrically
            # equivalent vectors
            idx = np.ones(v3.size, dtype=int) * -1

            # Loop over initial vectors
            l_accum = 0
            for i in range(n_v):
                # Unique vectors among those symmetrically equivalent
                vi = v2[:, i].unique()
                l = vi.size  # Multiplicity
                v3[i, :l] = vi  # Insert only the unique ones

                # Multiplicity of this initial vector
                multiplicity[i] = l

                # Index into `self` for the unique, symmetrically
                # equivalent vectors
                idx[l_accum : l_accum + l] = i
                l_accum += l

            # Remove entries into `v3` and `idx` not used
            non_zero = np.sum(np.abs(v3.data), axis=-1) != 0
            v2 = v3[non_zero]
            idx = idx[: np.sum(non_zero)]

        v2 = v2.flatten()

        # Carry over crystal structure and coordinate format
        m = self.__class__(xyz=v2.data, phase=self.phase)
        m.coordinate_format = self.coordinate_format

        if return_multiplicity and return_index:
            return m, multiplicity, idx
        elif return_multiplicity and not return_index:
            return m, multiplicity
        elif not return_multiplicity and return_index:
            return m, idx
        else:
            return m

    def angle_with(
        self,
        other: Miller,
        use_symmetry: bool = False,
        degrees: bool = False,
    ) -> np.ndarray:
        """Return the angles between these vectors and the other vectors
        possibly using symmetrically equivalent vectors to find the
        smallest angle under symmetry.

        Vectors must have compatible shapes, and be in the same space
        (direct or reciprocal) and crystal reference frames.

        Parameters
        ----------
        other
            Other vectors.
        use_symmetry
            Whether to consider equivalent vectors to find the smallest
            angle under symmetry. Default is ``False``.
        degrees
            If ``True``, the given angles are returned in degrees.
            Default is ``False``.

        Returns
        -------
        angles
            Angles in radians (``degrees=False``)  or degrees
            (``degrees=True``).
        """
        self._compatible_with(other, raise_error=True)

        if use_symmetry:
            other2 = other.symmetrise(unique=True)
            cosines = self.dot_outer(other2) / (
                self.norm[..., np.newaxis] * other2.norm[np.newaxis, ...]
            )
            cosines = np.round(cosines, 12)
            angles = np.min(np.arccos(cosines), axis=-1)
        else:
            angles = super().angle_with(other)

        if degrees:
            angles = np.rad2deg(angles)

        return angles

    def cross(self, other: Miller):
        """Return the cross products of the vectors with the other
        vectors, which is considered the zone axes between the vectors.

        Parameters
        ----------
        other
            Other vectors, which must be in the same space (direct or
            reciprocal) and have the same crystal reference frame.

        Returns
        -------
        m
            Vectors in reciprocal (direct) space if direct (reciprocal)
            vectors are crossed.
        """
        self._compatible_with(other, raise_error=True)
        new_fmt = dict(hkl="uvw", uvw="hkl", hkil="UVTW", UVTW="hkil")
        m = self.__class__(xyz=super().cross(other).data, phase=self.phase)
        m.coordinate_format = new_fmt[self.coordinate_format]
        return m

    def dot(self, other: Miller) -> np.ndarray:
        """Return the dot products of the vectors and the other vectors.

        Parameters
        ----------
        other
            Other vectors, which must be in the same space (direct or
            reciprocal) and have the same crystal reference frame.

        Returns
        -------
        dot_products
            Dot products.
        """
        self._compatible_with(other, raise_error=True)
        return super().dot(other)

    def dot_outer(self, other: Miller) -> np.ndarray:
        """Return the outer dot products of the vectors and the other
        vectors.

        Parameters
        ----------
        other
            Other vectors, which must be in the same space (direct or
            reciprocal) and have the same crystal reference frame.

        Returns
        -------
        dot_products
            Dot products.
        """
        self._compatible_with(other, raise_error=True)
        return super().dot_outer(other)

    def flatten(self) -> Miller:
        """Return the flattened vectors.

        Returns
        -------
        m
            Flattened vectors.
        """
        m = self.__class__(xyz=super().flatten().data, phase=self.phase)
        m.coordinate_format = self.coordinate_format
        return m

    def transpose(self, *axes: Optional[int]) -> Miller:
        """Return a new instance with the data transposed.

        The order may be undefined if :attr:`ndim` is originally 2. In
        this case the first two dimensions are transposed.

        Parameters
        ----------
        axes
            Transposed axes order. Only navigation axes need to be
            defined. May be undefined if self only contains two
            navigation dimensions.

        Returns
        -------
        m
            New transposed Miller instance of the original instance.
        """
        m = self.__class__(xyz=super().transpose(*axes).data, phase=self.phase)
        m.coordinate_format = self.coordinate_format
        return m

    def get_nearest(self, *args) -> NotImplemented:
        """NotImplemented."""
        return NotImplemented

    def mean(self, use_symmetry: bool = False) -> Miller:
        """Return the mean vector of the set of vectors.

        Parameters
        ----------
        use_symmetry
            Not implemented yet.

        Returns
        -------
        m
            Mean vector.
        """
        # TODO: Allow using symmetry by projecting to fundamental sector
        if use_symmetry:
            return NotImplemented
        m = self.__class__(xyz=super().mean().data, phase=self.phase)
        m.coordinate_format = self.coordinate_format
        return m

    def reshape(self, *shape: Union[int, tuple]) -> Miller:
        """Return a new instance with the vectors reshaped.

        Parameters
        ----------
        *shape
            The new shape as one or more integers or as a tuple.

        Returns
        -------
        m
            New instance.
        """
        m = self.__class__(xyz=super().reshape(*shape).data, phase=self.phase)
        m.coordinate_format = self.coordinate_format
        return m

    def unique(
        self, use_symmetry: bool = False, return_index: bool = False
    ) -> Union[Miller, Tuple[Miller, np.ndarray]]:
        """Unique vectors in ``self``.

        Parameters
        ----------
        use_symmetry
            Whether to consider equivalent vectors to compute the unique
            vectors. Default is ``False``.
        return_index
            Whether to return the indices of the (flattened) data where
            the unique entries were found. Default is ``False``.

        Returns
        -------
        m
            Flattened unique vectors.
        idx
            Indices of the unique data in the (flattened) array.
        """
        out = super().unique(return_index=return_index)
        if return_index:
            v, idx = out
        else:
            v = out

        if use_symmetry:
            operations = self.phase.point_group
            n_v = v.size
            v2 = operations.outer(v).flatten().reshape(n_v, operations.size)
            data = v2.data.round(10)
            data_sorted = np.zeros_like(data)
            for i in range(n_v):
                a = data[i]
                order = np.lexsort(a.T)  # Sort by column 1, 2, then 3
                data_sorted[i] = a[order]
            _, idx = np.unique(data_sorted, return_index=True, axis=0)
            v = v[idx[::-1]]

        m = self.__class__(xyz=v.data, phase=self.phase)
        m.coordinate_format = self.coordinate_format
        if return_index:
            return m, idx
        else:
            return m

    def in_fundamental_sector(
        self, symmetry: Optional["orix.quaternion.Symmetry"] = None
    ) -> Miller:
        """Project Miller indices to a symmetry's fundamental sector
        (inverse pole figure).

        This projection is taken from MTEX'
        :code:`project2FundamentalRegion`.

        Parameters
        ----------
        symmetry
            Symmetry with a fundamental sector, possibly not equal to
            :attr:`~orix.crystal_map.Phase.point_group`. If not given,
            ``point_group`` is used if valid, otherwise an error is
            raised.

        Returns
        -------
        m
            Vectors within the fundamental sector.

        Examples
        --------
        >>> from orix.crystal_map import Phase
        >>> from orix.quaternion.symmetry import D6h
        >>> from orix.vector import Miller
        >>> t = Miller(uvw=(-1, 1, 0), phase=Phase(point_group="m-3m"))
        >>> t.in_fundamental_sector()
        Miller (1,), point group m-3m, uvw
        [[1. 0. 1.]]
        >>> t.in_fundamental_sector(D6h)
        Miller (1,), point group m-3m, uvw
        [[1.366 0.366 0.   ]]
        """
        if symmetry is None:
            symmetry = self.phase.point_group
            if symmetry is None:
                raise ValueError(
                    "`symmetry` must be passed or `self.phase.point_group` must be a "
                    "`Symmetry` with a `Symmetry.fundamental_sector`"
                )
        v = Vector3d(self.data).in_fundamental_sector(symmetry)
        m = self.__class__(xyz=v.data, phase=self.phase)
        m.coordinate_format = self.coordinate_format
        return m

    # -------------------- Other private methods --------------------- #

    def _compatible_with(self, other: Miller, raise_error: bool = False) -> bool:
        """Whether ``self`` and ``other`` are the same (the same crystal
        lattice and symmetry) with vectors in the same space.

        Parameters
        ----------
        other
            Another vector instance.
        raise_error
            Whether to raise a ``ValueError`` if the instances are
            incompatible (default is ``False``).

        Returns
        -------
        compatible
            Whether they are compatible.
        """
        same_symmetry = self.phase.point_group == other.phase.point_group
        same_lattice = np.allclose(
            self.phase.structure.lattice.abcABG(),
            other.phase.structure.lattice.abcABG(),
        )
        same_space = self.space == other.space
        compatible = same_symmetry * same_lattice * same_space
        if not compatible and raise_error:
            raise ValueError(
                "The crystal lattices and symmetries must be the same, and the "
                "vector(s) must be in the same space"
            )
        else:
            return compatible


def _transform_space(
    v_in: np.ndarray, space_in: str, space_out: str, lattice: Lattice
) -> np.ndarray:
    r"""Convert vectors in a unit cell from one space to another.

    Parameters
    ----------
    v_in
        Input vectors.
    space_in, space_out
        ``"d"`` for direct (uvw), ``"r"`` for reciprocal (hkl) or
        ``"c"`` for cartesian (xyz).

    Returns
    -------
    v_out
        Output vectors.

    Notes
    -----
    Conversions between direct lattice vectors :math:`[uvw]`, reciprocal
    lattice vectors :math:`(hkl)` and Cartesian vectors
    :math:`(x, y, z)` using the structure matrix :math:`A` and the
    metric tensor :math:`g_{ij}`, where vectors are row vectors and
    matrices are row matrices:

    .. math::

        (x, y, z) = [uvw] \cdot \mathbf{A}
        (x, y, z) = (hkl) \cdot (\mathbf{A}^{-1})^T
        [uvw] = (x, y, z) \cdot \mathbf{A}^{-1}
        [uvw] = (hkl) \cdot g_{ij}^{-1}
        (hkl) = (x, y, z) \cdot A^T
        (hkl) = [uvw] \cdot g_{ij}
    """
    spaces = ["d", "r", "c"]
    if space_in not in spaces or space_out not in spaces:
        raise ValueError(f"`space_in` and `space_out` must be one of {spaces}")

    if space_in == space_out:
        v_out = np.copy(v_in)
    elif space_in == "d":
        if space_out == "c":
            # xyz = uvw * A
            v_out = np.matmul(v_in, lattice.base)
        else:
            # hkl = uvw * g_ij
            v_out = np.matmul(v_in, lattice.metrics)
    elif space_in == "r":
        if space_out == "c":
            # xyz = hkl * (A^-1)^T
            v_out = np.matmul(v_in, lattice.recbase.T)
        else:
            # uvw = hkl * g_ij^-1
            v_out = np.matmul(v_in, lattice.reciprocal().metrics)
    else:
        if space_out == "d":
            # uvw = xyz * A^-1
            v_out = np.dot(v_in, lattice.recbase)
        else:
            # hkl = xyz * ((A^-1)^T)^-1 = xyz * A^T
            v_out = np.matmul(v_in, lattice.base.T)

    return v_out


def _hkl2hkil(hkl: np.ndarray) -> np.ndarray:
    hkl = np.asarray(hkl)
    hkil = np.zeros(hkl.shape[:-1] + (4,))
    h = hkl[..., 0]
    k = hkl[..., 1]
    hkil[..., 0] = h
    hkil[..., 1] = k
    hkil[..., 2] = -(h + k)
    hkil[..., 3] = hkl[..., 2]
    return hkil


def _hkil2hkl(hkil: np.ndarray) -> np.ndarray:
    hkil = np.asarray(hkil)
    hkl = np.zeros(hkil.shape[:-1] + (3,))
    hkl[..., :2] = hkil[..., :2]
    hkl[..., 2] = hkil[..., 3]
    return hkl


def _check_hkil(hkil: np.ndarray):
    hkil = np.asarray(hkil)
    if not np.allclose(np.sum(hkil[..., :3], axis=-1), 0, atol=1e-4):
        raise ValueError(
            "The Miller-Bravais indices convention h + k + i = 0 is not satisfied"
        )


def _uvw2UVTW(uvw: np.ndarray, convention: Optional[str] = None) -> np.ndarray:
    uvw = np.asarray(uvw)
    UVTW = np.zeros(uvw.shape[:-1] + (4,))
    u = uvw[..., 0]
    v = uvw[..., 1]
    # DeGraef: U = (2u - v) / 3, V = (2v - u) / 3, T = -(u + v) / 3, W = w
    UVTW[..., 0] = (2 * u - v) / 3
    UVTW[..., 1] = (2 * v - u) / 3
    UVTW[..., 2] = -(u + v) / 3
    UVTW[..., 3] = uvw[..., 2]
    if convention is not None and convention.lower() == "mtex":
        # MTEX: U = 2u - v, V = 2v - u, T = -(u + v), W = 3w
        UVTW *= 3
    return UVTW


def _UVTW2uvw(UVTW: np.ndarray, convention: Optional[str] = None) -> np.ndarray:
    UVTW = np.asarray(UVTW)
    uvw = np.zeros(UVTW.shape[:-1] + (3,))
    # DeGraef: u = 2U + V, v = 2V + U, w = W
    U = UVTW[..., 0]
    V = UVTW[..., 1]
    uvw[..., 0] = 2 * U + V
    uvw[..., 1] = U + 2 * V
    uvw[..., 2] = UVTW[..., 3]
    if convention is not None and convention.lower() == "mtex":
        # MTEX: u = 2U + V, v = 2V + U, w = W / 3
        uvw /= 3
    return uvw


def _check_UVTW(UVTW: np.ndarray):
    UVTW = np.asarray(UVTW)
    if not np.allclose(np.sum(UVTW[..., :3], axis=-1), 0, atol=1e-4):
        raise ValueError(
            "The Miller-Bravais indices convention U + V + T = 0 is not satisfied"
        )


def _get_indices_from_highest(
    highest_indices: Union[list, tuple, np.ndarray]
) -> np.ndarray:
    """Return a list of coordinates from a set of highest indices.

    Parameters
    ----------
    highest_indices
        Highest indices to consider.

    Returns
    -------
    indices
        An array of indices sorted from positive to negative in the
        first column.
    """
    highest_indices = np.asarray(highest_indices)
    if not np.all(highest_indices >= 0) or np.all(highest_indices == 0):
        raise ValueError(
            f"All indices {highest_indices} must be positive with at least one non-zero"
        )
    index_ranges = [np.arange(-i, i + 1) for i in highest_indices]
    indices = np.asarray(list(product(*index_ranges)), dtype=int)
    indices = indices[~np.all(indices == 0, axis=1)]  # Remove (000)
    indices = indices[::-1]  # Make e.g. (111) first instead of (-1-1-1)
    return indices


def _get_highest_hkl(lattice: Lattice, min_dspacing: float = 0.05) -> np.ndarray:
    """Return the highest Miller indices hkl of the plane with a direct
    space interplanar spacing closest to a threshold.

    Parameters
    ----------
    lattice
        Crystal lattice.
    min_dspacing
        Smallest interplanar spacing to consider. Default is 0.05.

    Returns
    -------
    highest_hkl
        Highest Miller indices.
    """
    highest_hkl = np.ones(3, dtype=int)
    for i in range(3):
        hkl = np.zeros(3)
        hkl[i] = 1
        d = 1 / lattice.rnorm(hkl)
        while d > min_dspacing:
            hkl[i] += 1
            d = 1 / lattice.rnorm(hkl)
        highest_hkl[i] = hkl[i] - 1
    return highest_hkl


def _round_indices(
    indices: Union[list, tuple, np.ndarray], max_index: int = 12
) -> np.ndarray:
    """Round a set of index triplet (Miller) or quartet (Miller-Bravais)
    to the *closest* smallest integers.

    Adopted from MTEX' :code:`Miller.round` function.

    Parameters
    ----------
    indices
        Set of index triplet(s) or quartet(s) to round.
    max_index
        Maximum integer index to round to, by default 12.

    Return
    ------
    new_indices
        Integer array of rounded set of index triplet(s) or quartet(s).
    """
    # Allow list and tuple input (and don't overwrite `indices`)
    idx = np.asarray(indices)

    # Flatten and remove redundant third index if Miller-Bravais
    n_idx = idx.shape[-1]  # 3 or 4
    idx_flat = np.reshape(idx, (-1, n_idx))
    if n_idx == 4:
        idx_flat = idx_flat[..., [0, 1, 3]]

    # Get number of sets, max. index per set, and all possible integer
    # multipliers between 1 and `max_index`
    n_sets = idx_flat.size // 3
    max_per_set = np.max(np.abs(idx_flat), axis=-1)
    multipliers = np.arange(1, max_index + 1)

    # Divide by highest index, repeat array `max_index` number of times,
    # and multiply with all multipliers
    idx_scaled = (
        np.broadcast_to(idx_flat / max_per_set[..., np.newaxis], (max_index, n_sets, 3))
        * multipliers[..., np.newaxis, np.newaxis]
    )

    # Find the most suitable multiplier per set, which gives the
    # smallest error between the initial set and the scaled and rounded
    # set
    error = 1e-7 * np.round(
        1e7
        * np.sum((idx_scaled - np.round(idx_scaled)) ** 2, axis=-1)
        / np.sum(idx_scaled**2, axis=-1)
    )
    idx_min_error = np.argmin(error, axis=0)
    multiplier = (idx_min_error + 1) / max_per_set

    # Reshape `multiplier` to match indices shape
    multiplier = multiplier.reshape(idx.shape[:-1])[..., np.newaxis]

    # Finally, multiply each set with their most suitable multiplier,
    # and round
    new_indices = np.round(multiplier * idx).astype(int)

    return new_indices



r"""
=======================================
Misorientation from aligning directions
=======================================

This example demonstrates how to use
:meth:`~orix.quaternion.Misorientation.from_align_vectors` to estimate a misorientation
:math:`M` from two sets of aligned crystal directions :math:`t_{1,2}`, one set in each
crystal reference frame :math:`(e_1, e_2, e_3)`.
"""

from diffpy.structure import Lattice, Structure
import matplotlib.pyplot as plt
import numpy as np

from orix.crystal_map import Phase
from orix.quaternion import Misorientation, Orientation
from orix.vector import Miller

plt.rcParams.update({"figure.figsize": (5, 5), "lines.markersize": 8})

# Specify a cubic and an hexagonal crystal structures and symmetries
phase1 = Phase(
    point_group="m-3m",
    structure=Structure(lattice=Lattice(1, 1, 1, 90, 90, 90)),
)
phase2 = Phase(
    point_group="6/mmm",
    structure=Structure(lattice=Lattice(1, 1, 2, 90, 90, 120)),
)

# Specify one orientation per crystal
O_cubic = Orientation.from_axes_angles(
    [1, 1, 1], -60, symmetry=phase1.point_group, degrees=True
)
O_hex = Orientation.from_axes_angles(
    [1, 3, 2], -90, symmetry=phase2.point_group, degrees=True
)

# Get the reference misorientation (goal). Misorientations are obtained
# from the right, so: crystal 1 -> sample -> crystal 2
M_ref = Misorientation(O_hex * ~O_cubic, symmetry=(O_cubic.symmetry, O_hex.symmetry))

# Specify two directions in the first crystal
t_cubic = Miller(uvw=[[1, 1, 1], [0, 0, 1]], phase=phase1)

# Express the same directions with respect to the second crystal [in the
# Cartesian reference frame, (e1, e2, e3)]
v_hex = Miller(xyz=(M_ref * t_cubic).data, phase=phase2)

# Add some randomness to the second crystal directions (0 error
# magnitude gives exact result)
error_magnitude = 0.1
v_err = Miller(xyz=np.random.normal(0, error_magnitude, 3), phase=phase2)
v_hex_err = Miller(xyz=(v_hex + v_err).data, phase=phase2)
angle_err = v_hex_err.angle_with(v_hex, degrees=True).mean()
print("Vector angular deviation [deg]: ", angle_err)

# Get the misorientation that aligns the directions in the first crystal
# with those in the second crystal
M_new, err = Misorientation.from_align_vectors(v_hex_err, t_cubic, return_rmsd=True)
print("Error distance: ", err)

# Plot the two directions in the (unrotated) first crystal's reference
# frame as open circles
fig = t_cubic.scatter(
    c="none",
    ec=["r", "b"],
    grid=True,
    axes_labels=["e1", "e2"],
    return_figure=True,
    figure_kwargs={"layout": "tight"},
)

# Plot the two directions in the second crystal with respect to the
# first crystal's axes
(~M_ref * v_hex_err).scatter(c=["r", "b"], figure=fig)



"""
=================
Zoom region inset
=================

This example shows how to add a region in the stereographic projection as a zoomed
inset inside another stereographic projection, following the procedure in `this
Matplotlib example
<https://matplotlib.org/stable/gallery/subplots_axes_and_figures/zoom_inset_axes.html>`__.
"""

import matplotlib.pyplot as plt

from orix import plot, projections, sampling
from orix.vector import Vector3d

# Sample some orientations
v = sampling.sample_S2(2)
v_ref = Vector3d([1, 1, 1])
v2 = v[v_ref.angle_with(v, degrees=True) < 10]

# Plot them in the stereographic projection with a grid resolution of 10
# degrees
fig = plt.figure(figsize=(5, 5))
ax = fig.add_subplot(111, projection="stereographic")
ax.stereographic_grid(True)
ax.set_labels("X", "Y", None, fontsize=20)
ax.scatter(v2)
ax.scatter(v_ref, c="r")

# Define some vectors describing the x/y extent of the zoomed inset
# region and get their stereographic coordinates (X, Y)
v_inset = Vector3d.from_polar(azimuth=[45, 45], polar=[36, 71], degrees=True)
stereo = projections.StereographicProjection()
x_inset, y_inset = stereo.vector2xy(v_inset)

# The zoomed inset rectangle origin (x, y), width and height
rect = [0.15, 0.2, 0.43, 0.43]

# Add a new stereographic projection axis and zoom in
ax_inset = ax.inset_axes(rect, projection="stereographic")
ax_inset.set(
    xlim=(x_inset.min(), x_inset.max()),
    ylim=(y_inset.min(), y_inset.max()),
)

# Add a grid of 2 degrees resolution and re-plot the vectors
ax_inset.stereographic_grid(True, 2, 2)
ax_inset.scatter(v2)
ax_inset.scatter(v_ref, c="r")

# Add lines indicating the inset zoom
ax.indicate_inset_zoom(ax_inset, edgecolor="k")

# Add border to the inset region
for spine in ax_inset.spines.values():
    spine.set_visible(True)

# The inset axis is not compatible with fig.tight_layout(), so we set
# the origin and width and height manually
fig.subplots_adjust(0, 0, 1, 1)



# -*- coding: utf-8 -*-
# Copyright 2018-2024 the orix developers
#
# This file is part of orix.
#
# orix is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# orix is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with orix.  If not, see <http://www.gnu.org/licenses/>.

import numpy as np
import pytest

from orix.plot import IPFColorKeyTSL
from orix.plot.direction_color_keys._util import polar_coordinates_in_sector
from orix.quaternion import symmetry
from orix.vector import Vector3d


class TestDirectionColorKeyTSL:
    def test_direction2color(self):
        ckey_oh = IPFColorKeyTSL(symmetry.Oh)
        ckey_oh_direction = ckey_oh.direction_color_key
        assert repr(ckey_oh_direction) == "DirectionColorKeyTSL, symmetry m-3m"

    def test_triclinic(self):
        # Get RGB colors for C1. Will never reach from IPFColorKeyTSL
        # since Ci (-1) is the Laue group of C1.
        sector = symmetry.C1.fundamental_sector
        rgb2 = polar_coordinates_in_sector(sector, Vector3d.xvector())
        assert rgb2[0].size == rgb2[1].size == 0

    @pytest.mark.parametrize(
        "symmetry, expected_shape",
        [
            [symmetry.C1, (2000, 2000, 4)],
            [symmetry.C2, (1000, 2000, 4)],
            [symmetry.D6, (500, 1000, 4)],
            [symmetry.Oh, (367, 415, 4)],
            [symmetry.Th, (415, 415, 4)],
        ],
    )
    def test_rgba_grid_shape(self, symmetry, expected_shape):
        ckey = IPFColorKeyTSL(symmetry)
        ckey_direction = ckey.direction_color_key
        rgb_grid = ckey_direction._create_rgba_grid()
        assert isinstance(rgb_grid, np.ndarray)
        assert rgb_grid.shape == expected_shape

    @pytest.mark.parametrize(
        "symmetry, expected_xlim, expected_ylim",
        [
            [symmetry.C1, (-1.0, 1.0), (-1.0, 1.0)],
            [symmetry.C2, (-1.0, 1.0), (0.0, 1.0)],
            [symmetry.D6, (0.0, 1.0), (0.0, 0.5)],
            [symmetry.Oh, (0.0, 0.414), (0.0, 0.366)],
            [symmetry.Th, (0.0, 0.414), (0.0, 0.414)],
        ],
    )
    def test_rgba_grid_limits(self, symmetry, expected_xlim, expected_ylim):
        ckey = IPFColorKeyTSL(symmetry)
        ckey_direction = ckey.direction_color_key
        (
            _,
            (x_lim, y_lim),
        ) = ckey_direction._create_rgba_grid(return_extent=True)
        (x_min, x_max), (y_min, y_max) = x_lim, y_lim
        assert round(x_min, 3) == round(expected_xlim[0], 3)
        assert round(x_max, 3) == round(expected_xlim[1], 3)
        assert round(y_min, 3) == round(expected_ylim[0], 3)
        assert round(y_max, 3) == round(expected_ylim[1], 3)

    @pytest.mark.parametrize(
        "symmetry",
        [symmetry.C2, symmetry.D6, symmetry.Oh],
    )
    def test_rgba_grid_alpha(self, symmetry):
        ckey = IPFColorKeyTSL(symmetry)
        ckey_direction = ckey.direction_color_key
        rgba_grid = ckey_direction._create_rgba_grid()
        # test invalid points have alpha = 0
        assert (rgba_grid[0, 0] == (1, 1, 1, 0)).all()
        rgba_grid = ckey_direction._create_rgba_grid(alpha=0.5)
        valid = (rgba_grid != (1, 1, 1, 0)).all(axis=-1)
        assert (rgba_grid[valid][..., -1] == 0.5).all()



# Configuration file for the Sphinx documentation app.
# See the documentation for a full list of configuration options:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

from datetime import datetime
import inspect
import os
from os.path import dirname, relpath
import re
import sys

from numpydoc.docscrape_sphinx import SphinxDocString

import orix
from orix import data

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
# sys.path.insert(0, os.path.abspath("."))
sys.path.append("../")

project = "orix"
copyright = f"2018-{str(datetime.now().year)}, {orix.__author__}"
author = orix.__author__
release = orix.__version__

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "matplotlib.sphinxext.plot_directive",
    "nbsphinx",
    "sphinxcontrib.bibtex",
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.doctest",
    "sphinx.ext.imgconverter",
    "sphinx.ext.intersphinx",
    "sphinx.ext.linkcode",
    "sphinx.ext.mathjax",
    "sphinx_codeautolink",
    "sphinx_copybutton",
    "sphinx_design",
    "sphinx_gallery.gen_gallery",
    "numpydoc",  # Must be loaded after autodoc
]

# Create links to references within orix's documentation to these packages
intersphinx_mapping = {
    "black": ("https://black.readthedocs.io/en/stable", None),
    "coverage": ("https://coverage.readthedocs.io/en/latest", None),
    "dask": ("https://docs.dask.org/en/latest", None),
    "defdap": ("https://defdap.readthedocs.io/en/latest", None),
    "diffpy.structure": ("https://www.diffpy.org/diffpy.structure", None),
    "diffsims": ("https://diffsims.readthedocs.io/en/latest", None),
    "h5py": ("https://docs.h5py.org/en/stable", None),
    "kikuchipy": ("https://kikuchipy.org/en/latest", None),
    "matplotlib": ("https://matplotlib.org/stable", None),
    "nbsphinx": ("https://nbsphinx.readthedocs.io/en/latest", None),
    "nbval": ("https://nbval.readthedocs.io/en/latest", None),
    "numpy": ("https://numpy.org/doc/stable", None),
    "numpydoc": ("https://numpydoc.readthedocs.io/en/latest", None),
    "pytest": ("https://docs.pytest.org/en/stable", None),
    "python": ("https://docs.python.org/3", None),
    "pyxem": ("https://pyxem.readthedocs.io/en/latest", None),
    "readthedocs": ("https://docs.readthedocs.io/en/stable", None),
    "scipy": ("https://docs.scipy.org/doc/scipy", None),
    "sklearn": ("https://scikit-learn.org/stable", None),
    "sphinx": ("https://www.sphinx-doc.org/en/master", None),
    "sphinx-gallery": ("https://sphinx-gallery.github.io/stable", None),
}

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = [
    "build",
    "Thumbs.db",
    ".DS_Store",
    # Suppress warnings from Sphinx regarding "duplicate source files":
    # https://github.com/executablebooks/MyST-NB/issues/363#issuecomment-1682540222
    "examples/*/*.ipynb",
    "examples/*/*.py",
]

# HTML theming: pydata-sphinx-theme
# https://pydata-sphinx-theme.readthedocs.io
html_theme = "pydata_sphinx_theme"
html_theme_options = {
    "github_url": "https://github.com/pyxem/orix",
    "header_links_before_dropdown": 6,
    "logo": {"alt_text": project, "text": project},
    "navigation_with_keys": True,
    "show_toc_level": 2,
    "use_edit_page_button": True,
}
html_context = {
    "github_user": "pyxem",
    "github_repo": "orix",
    "github_version": "develop",
    "doc_path": "doc",
}
html_static_path = ["_static"]
html_css_files = ["custom.css"]

# Syntax highlighting
pygments_style = "friendly"

# Logo
html_logo = "_static/img/orix_logo.png"
html_favicon = "_static/img/orix_logo.png"

# -- nbsphinx
# https://nbsphinx.readthedocs.io
# Taken from nbsphinx' own nbsphinx configuration file, with slight
# modifications to point nbviewer and Binder to the GitHub develop
# branch links when the documentation is launched from a kikuchipy
# version with "dev" in the version
if "dev" in orix.__version__:
    release_version = "develop"
else:
    release_version = "v" + orix.__version__
# This is processed by Jinja2 and inserted before each notebook
nbsphinx_prolog = (
    r"""
{% set docname = 'doc/' + env.doc2path(env.docname, base=None)[:-6] + '.ipynb' %}

.. raw:: html

    <style>a:hover { text-decoration: underline; }</style>

    <div class="admonition note">
      This page was generated from
      <a class="reference external" href="https://github.com/pyxem/orix/blob/"""
    + f"{release_version}"
    + r"""/{{ docname|e }}">{{ docname|e }}</a>.
      Interactive online version:
      <span style="white-space: nowrap;"><a href="https://mybinder.org/v2/gh/pyxem/orix/"""
    + f"{release_version}"
    + r"""?filepath={{ docname|e }}"><img alt="Binder badge" src="https://mybinder.org/badge_logo.svg" style="vertical-align:text-bottom"></a>.</span>
      <script>
        if (document.location.host) {
          $(document.currentScript).replaceWith(
            '<a class="reference external" ' +
            'href="https://nbviewer.jupyter.org/url' +
            (window.location.protocol == 'https:' ? 's/' : '/') +
            window.location.host +
            window.location.pathname.slice(0, -4) +
            'ipynb">View in <em>nbviewer</em></a>.'
          );
        }
      </script>
    </div>

.. raw:: latex

    \nbsphinxstartnotebook{\scriptsize\noindent\strut
    \textcolor{gray}{The following section was generated from
    \sphinxcode{\sphinxupquote{\strut {{ docname | escape_latex }}}} \dotfill}}
"""
)
# https://nbsphinx.readthedocs.io/en/0.8.0/never-execute.html
nbsphinx_execute = "auto"  # auto, always, never
nbsphinx_allow_errors = True
nbsphinx_execute_arguments = [
    "--InlineBackend.rc=figure.facecolor='w'",
    "--InlineBackend.rc=font.size=15",
]

# -- sphinxcontrib-bibtex
# https://sphinxcontrib-bibtex.readthedocs.io
bibtex_bibfiles = ["user/bibliography.bib"]
bibtex_reference_style = "author_year"

# -- sphinx-codeautolink
codeautolink_custom_blocks = {
    "python3": None,
    "pycon3": "sphinx_codeautolink.clean_pycon",
}


def linkcode_resolve(domain, info):
    """Determine the URL corresponding to Python object.
    This is taken from SciPy's conf.py:
    https://github.com/scipy/scipy/blob/develop/doc/source/conf.py.
    """
    if domain != "py":
        return None

    modname = info["module"]
    fullname = info["fullname"]

    submod = sys.modules.get(modname)
    if submod is None:
        return None

    obj = submod
    for part in fullname.split("."):
        try:
            obj = getattr(obj, part)
        except Exception:
            return None

    # Use the original function object if it is wrapped.
    obj = getattr(obj, "__wrapped__", obj)

    try:
        fn = inspect.getsourcefile(obj)
    except Exception:
        fn = None
    if not fn:
        try:
            fn = inspect.getsourcefile(sys.modules[obj.__module__])
        except Exception:
            fn = None
    if not fn:
        return None

    try:
        source, lineno = inspect.getsourcelines(obj)
    except Exception:
        lineno = None

    if lineno:
        linespec = "#L%d-L%d" % (lineno, lineno + len(source) - 1)
    else:
        linespec = ""

    startdir = os.path.abspath(os.path.join(dirname(orix.__file__), ".."))
    fn = relpath(fn, start=startdir).replace(os.path.sep, "/")

    if fn.startswith("orix/"):
        m = re.match(r"^.*dev0\+([a-f\d]+)$", orix.__version__)
        pre_link = "https://github.com/pyxem/orix/blob/"
        if m:
            return pre_link + "%s/%s%s" % (m.group(1), fn, linespec)
        elif "dev" in orix.__version__:
            return pre_link + "develop/%s%s" % (fn, linespec)
        else:
            return pre_link + "v%s/%s%s" % (orix.__version__, fn, linespec)
    else:
        return None


# -- Copy button customization (taken from PyVista)
# Exclude traditional Python prompts from the copied code
copybutton_prompt_text = r">>> ?|\.\.\. "
copybutton_prompt_is_regexp = True


# -- sphinx.ext.autodoc
# https://www.sphinx-doc.org/en/master/usage/extensions/autodoc.html
autosummary_ignore_module_all = False
autosummary_imported_members = True
autodoc_typehints_format = "short"
autodoc_default_options = {
    "show-inheritance": True,
}


# -- numpydoc
# https://numpydoc.readthedocs.io
numpydoc_show_class_members = False
numpydoc_use_plots = True
numpydoc_xref_param_type = True
# fmt: off
numpydoc_validation_checks = {
    "all",   # All but the following:
    "ES01",  # Not all docstrings need an extend summary
    "EX01",  # Examples: Will eventually enforce
    "GL01",  # Contradicts numpydoc examples
    "GL02",  # Appears to be broken?
    "GL07",  # Appears to be broken?
    "GL08",  # Methods can be documented in super class
    "PR01",  # Parameters can be documented in super class
    "PR02",  # Properties with setters might have docstrings w/"Returns"
    "PR04",  # Doesn't seem to work with type hints?
    "RT01",  # Abstract classes might not have return sections
    "SA01",  # Not all docstrings need a "See Also"
    "SA04",  # "See Also" section does not need descriptions
    "SS06",  # Not possible to make all summaries one line
    "YD01",  # Yields: No plan to enforce
}
# fmt: on


# -- matplotlib.sphinxext.plot_directive
# https://matplotlib.org/stable/api/sphinxext_plot_directive_api.html
plot_formats = ["png"]
plot_html_show_source_link = False
plot_html_show_formats = False
plot_include_source = True


def _str_examples(self):
    examples_str = "\n".join(self["Examples"])
    if (
        self.use_plots
        and (
            re.search(r"\b(.plot)\b", examples_str)
            or re.search(r"\b(.plot_map)\b", examples_str)
            or re.search(r"\b(.imshow)\b", examples_str)
        )
        and "plot::" not in examples_str
    ):
        out = []
        out += self._str_header("Examples")
        out += [".. plot::", ""]
        out += self._str_indent(self["Examples"])
        out += [""]
        return out
    else:
        return self._str_section("Examples")


SphinxDocString._str_examples = _str_examples


# -- Sphinx-Gallery
# https://sphinx-gallery.github.io
sphinx_gallery_conf = {
    "backreferences_dir": "reference/generated",
    "doc_module": ("orix",),
    "examples_dirs": "../examples",
    "filename_pattern": "^((?!sgskip).)*$",
    "gallery_dirs": "examples",
    "reference_url": {"orix": None},
    "run_stale_examples": True,
    "show_memory": True,
}
autosummary_generate = True


# Download example datasets prior to building the docs
print("[orix] Downloading example datasets (if not found in the cache)")
_ = data.sdss_ferrite_austenite(allow_download=True)
_ = data.sdss_austenite(allow_download=True)
_ = data.ti_orientations(allow_download=True)


def skip_member(app, what, name, obj, skip, options):
    """Exclude objects not defined within orix from the API reference.

    This ensures inherited members from Matplotlib axes extensions are
    excluded from the reference. We could exclude inherited members
    all together in the Sphinx Jinja2 template. But, this would mean
    classes such as Rotation would have to "overwrite" all methods and
    properties from Quaternion for these members to be listed in the
    API reference of Rotation. Instead, we allow inherited members, but
    try our best to skip members coming from outside orix here.
    """
    if what in ["attribute", "property"]:
        obj_module = inspect.getmodule(getattr(obj, "fget", None))
    else:
        obj_module = inspect.getmodule(obj)
    return "orix" not in getattr(obj_module, "__name__", [])


def setup(app):
    app.connect("autodoc-skip-member", skip_member)



"""
=============================
Inverse pole density function
=============================

This example shows how to plot an inverse pole density function (IPDF)
:cite:`rohrer2004distribution` to inspect the distribution of crystal directions
:math:`t` pointing in some direction in the sample reference frame :math:`(x, y, z)`.
"""

import matplotlib.pyplot as plt

from orix import data, plot
from orix.vector import Vector3d

xmap = data.sdss_ferrite_austenite(allow_download=True)
print(xmap)

# Extract orientations, O
pg_m3m = xmap.phases[1].point_group.laue
O_fe = xmap["ferrite"].orientations
O_au = xmap["austenite"].orientations

# Some sample direction, v
v = Vector3d([0, 0, 1])
v_title = "Z"

# Rotate sample direction v into every crystal orientation O
t_fe = O_fe * v
t_au = O_au * v

# Set IPDF range
vmin, vmax = (0, 3)

subplot_kw = {"projection": "ipf", "symmetry": pg_m3m, "direction": v}
fig = plt.figure(figsize=(9, 8))

ax0 = fig.add_subplot(221, **subplot_kw)
ax0.scatter(O_fe, alpha=0.05)
_ = ax0.set_title(f"Ferrite, {v_title}")

ax1 = fig.add_subplot(222, **subplot_kw)
ax1.scatter(O_au, alpha=0.05)
_ = ax1.set_title(f"Austenite, {v_title}")

ax2 = fig.add_subplot(223, **subplot_kw)
ax2.pole_density_function(t_fe, vmin=vmin, vmax=vmax)
_ = ax2.set_title(f"Ferrite, {v_title}")

ax3 = fig.add_subplot(224, **subplot_kw)
ax3.pole_density_function(t_au, vmin=vmin, vmax=vmax)
_ = ax3.set_title(f"Austenite, {v_title}")



# -*- coding: utf-8 -*-
# Copyright 2018-2024 the orix developers
#
# This file is part of orix.
#
# orix is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# orix is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with orix.  If not, see <http://www.gnu.org/licenses/>.

"""Helper functions and classes for managing orix."""


import functools
import inspect
import warnings

import numpy as np


class deprecated:
    """Decorator to mark deprecated functions or properties with an
    informative warning.

    Adapted from
    `scikit-image
    <https://github.com/scikit-image/scikit-image/blob/main/skimage/_shared/utils.py#L297>`_
    and `matplotlib
    <https://github.com/matplotlib/matplotlib/blob/master/lib/matplotlib/_api/deprecation.py#L122>`_.
    """

    def __init__(
        self,
        since,
        removal=None,
        object_type="function",
        alternative=None,
    ):
        """Visible deprecation warning.

        Parameters
        ----------
        since : str, int or float
            The release at which this API became deprecated.
        removal : str, int or float, optional
            The expected removal version.
        object_type : str, optional
            Type of the deprecated object, either "function" (default)
            or "property".
        alternative : str, optional
            An alternative API that the user may use in place of the
            deprecated API.
        """
        self.since = since
        self.alternative = alternative
        self.removal = removal
        self.object_type = object_type

    def __call__(self, func):
        # Wrap function or property to raise warning when called, and
        # add warning to docstring if function or property is deprecated

        object_type = self.object_type.lower()
        if object_type == "function":
            parentheses = "()"
        else:
            parentheses = ""
        if self.alternative is not None:
            alt_msg = f" Use `{self.alternative}{parentheses}` instead."
        else:
            alt_msg = ""
        if self.removal is not None:
            rm_msg = f" and will be removed in version {self.removal}"
        else:
            rm_msg = ""
        msg = (
            f"{object_type.capitalize()} `{func.__name__}{parentheses}` is deprecated"
            f"{rm_msg}.{alt_msg}"
        )

        @functools.wraps(func)
        def wrapped(*args, **kwargs):
            warnings.simplefilter(
                action="always", category=np.VisibleDeprecationWarning, append=True
            )
            func_code = func.__code__
            warnings.warn_explicit(
                message=msg,
                category=np.VisibleDeprecationWarning,
                filename=func_code.co_filename,
                lineno=func_code.co_firstlineno + 1,
            )
            return func(*args, **kwargs)

        # Modify docstring to display deprecation warning
        old_doc = inspect.cleandoc(func.__doc__ or "").strip("\n")
        notes_header = "\nNotes\n-----"
        new_doc = (
            f"[*Deprecated*] {old_doc}\n"
            f"{notes_header if notes_header not in old_doc else ''}\n"
            f".. deprecated:: {self.since}\n"
            f"   {msg.strip()}"  # Matplotlib uses three spaces
        )
        wrapped.__doc__ = new_doc

        return wrapped


class deprecated_argument:
    """Decorator to remove an argument from a function or method's
    signature.

    Adopted from
    `scikit-image
    <https://github.com/scikit-image/scikit-image/blob/main/skimage/_shared/utils.py#L115>`_.
    """

    def __init__(self, name, since, removal, alternative=None):
        self.name = name
        self.since = since
        self.removal = removal
        self.alternative = alternative

    def __call__(self, func):
        @functools.wraps(func)
        def wrapped(*args, **kwargs):
            if self.name in kwargs.keys():
                msg = (
                    f"Argument `{self.name}` is deprecated and will be removed in "
                    f"version {self.removal}. To avoid this warning, please do not use "
                    f"`{self.name}`. "
                )
                if self.alternative is not None:
                    msg += f"Use `{self.alternative}` instead. "
                msg += f"See the documentation of `{func.__name__}()` for more details."
                warnings.simplefilter(
                    action="always", category=np.VisibleDeprecationWarning, append=True
                )
                func_code = func.__code__
                warnings.warn_explicit(
                    message=msg,
                    category=np.VisibleDeprecationWarning,
                    filename=func_code.co_filename,
                    lineno=func_code.co_firstlineno + 1,
                )
            return func(*args, **kwargs)

        return wrapped



# -*- coding: utf-8 -*-
# Copyright 2018-2024 the orix developers
#
# This file is part of orix.
#
# orix is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# orix is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with orix.  If not, see <http://www.gnu.org/licenses/>.

"""Inverse pole figure plot inheriting from
:class:`~orix.plot.StereographicPlot` for plotting of
:class:`~orix.vector.Vector3d`, typically parallel to sample directions,
rotated by orientations.
"""

from typing import Any, Dict, List, Optional, Tuple, Union

import matplotlib.axes as maxes
from matplotlib.figure import Figure
from matplotlib.patches import PathPatch
import matplotlib.projections as mprojections
import matplotlib.pyplot as plt
import numpy as np

from orix.crystal_map import Phase
from orix.measure import pole_density_function
from orix.plot.direction_color_keys.direction_color_key_tsl import DirectionColorKeyTSL
from orix.plot.stereographic_plot import ZORDER, StereographicPlot
from orix.quaternion import Orientation
from orix.quaternion.symmetry import C1, Symmetry
from orix.vector import Miller, Vector3d


class InversePoleFigurePlot(StereographicPlot):
    """Inverse pole figure plot of :class:`~orix.vector.Vector3d`, which
    is a stereographic plot for showing sample directions with respect
    to a crystal reference frame.

    Parameters
    ----------
    *args
        Arguments passed to :class:`~orix.plot.StereographicPlot`.
    symmetry
        Laue group symmetry of crystal to plot directions with. If not
        given, point group C1 (only identity rotation) is used.
    direction
        Sample direction to plot with respect to crystal directions. If
        not given, the out of plane direction, sample Z, is used.
    hemisphere
        Which hemisphere(s) to plot the vectors in. If not given,
        ``"upper"`` is used. Options are ``"upper"``, ``"lower"`` and
        ``"both"``, which plots two projections side by side.
    **kwargs
        Keyword arguments passed to
        :class:`~orix.plot.StereographicPlot`.
    """

    name = "ipf"

    def __init__(
        self,
        *args: Any,
        symmetry: Optional[Symmetry] = None,
        direction: Optional[Vector3d] = None,
        hemisphere: Optional[str] = None,
        **kwargs: Any,
    ) -> None:
        """Create an inverse pole figure axis for plotting
        :class:`~orix.vector.Vector3d`.
        """
        super().__init__(*args, **kwargs)

        if hemisphere is not None:
            self.hemisphere = hemisphere

        if direction is None:
            direction = Vector3d.zvector()
        self._direction = direction

        if symmetry is None:
            symmetry = C1
        self._symmetry = symmetry

        self.restrict_to_sector(self._symmetry.fundamental_sector)

        self._add_crystal_direction_labels()

    @property
    def _edge_patch(self) -> PathPatch:
        """Easy access to the fundamental sector border patch."""
        patches = self.patches
        return patches[self._has_collection(label="sa_sector", collections=patches)[1]]

    def pole_density_function(
        self,
        *args: Union[np.ndarray, Vector3d],
        resolution: float = 0.25,
        sigma: float = 5,
        log: bool = False,
        colorbar: bool = True,
        weights: Optional[np.ndarray] = None,
        **kwargs: Any,
    ):
        """Compute the Inverse Pole Density Function (IPDF) of vectors
        in the stereographic projection.

        The PDF is computed within the fundamental sector of the point
        group symmetry. See :cite:`rohrer2004distribution`.

        Parameters
        ----------
        *args
            Vector(s), or azimuth and polar angles of the vectors, the
            latter passed as two separate arguments.
        resolution
            The angular resolution of the sampling grid in degrees.
            Default value is 0.25.
        sigma
            The angular resolution of the applied broadening in degrees.
            Default value is 5.
        log
            If ``True`` the log(IPDF) is calculated. Default is
            ``True``.
        colorbar
            If ``True`` a colorbar is shown alongside the IPDF plot.
            Default is ``True``.
        weights
            The weights for the individual vectors. If not given, each
            vector is 1.
        **kwargs
            Keyword arguments passed to
            :meth:`matplotlib.axes.Axes.pcolormesh`.

        See Also
        --------
        measure.pole_density_function
        StereographicPlot.pole_density_function
        Vector3d.pole_density_function
        """
        hist, (x, y) = pole_density_function(
            *args,
            resolution=resolution,
            sigma=sigma,
            log=log,
            hemisphere=self.hemisphere,
            symmetry=self._symmetry,
            weights=weights,
        )

        new_kwargs = dict(zorder=ZORDER["mesh"], clip_on=True)
        updated_kwargs = {**kwargs, **new_kwargs}

        # plot mesh
        updated_kwargs.setdefault("cmap", "magma")
        # mpl.QuadMesh handles masked values by default
        pc = self.pcolormesh(x, y, hist, **updated_kwargs)

        if colorbar:
            label = "MRD"
            if log:
                label = f"log({label})"
            plt.colorbar(pc, label=label, ax=self)

    def scatter(
        self,
        *args: Union[Tuple[np.ndarray, np.ndarray], Orientation, Vector3d],
        **kwargs: Any,
    ) -> None:
        """A scatter plot of sample directions rotated by orientations,
        or orientations to rotate sample directions with.

        Parameters
        ----------
        *args
            Spherical coordinates (azimuth, polar), orientations, or
            vectors. If spherical coordinates are given, they are
            assumed to describe unit vectors. Vectors will be made into
            unit vectors if they aren't already. If orientations are
            passed, the crystal directions returned are the sample
            :attr:`direction` rotated by the orientations.
        **kwargs
            Keyword arguments passed to
            :meth:`matplotlib.axes.Axes.scatter`.

        See Also
        --------
        matplotlib.axes.Axes.scatter
        """
        crystal_directions = self._pretransform_input_ipf(args)
        super().scatter(crystal_directions, **kwargs)

    def show_hemisphere_label(self, **kwargs: Any) -> None:
        """Add a hemisphere label (``"upper"``/``"lower"``) to the upper
        left outside the inverse pole figure.

        Parameters
        ----------
        **kwargs
            Keyword arguments passed to
            :meth:`matplotlib.axes.Axes.text`.

        See Also
        --------
        hemisphere
        """
        new_kwargs = dict(ha="right", va="bottom")
        new_kwargs.update(kwargs)
        # TODO: Fix plotting of hemisphere labels for fundamental
        #  sectors with only two vertices on either side of equator (C3)
        x, y = self._edge_patch.get_path().vertices.T
        v = self._inverse_projection.xy2vector(np.min(x), np.max(y))
        self.text(v, s=self.hemisphere, **new_kwargs)

    def _add_crystal_direction_labels(self) -> None:
        """Add appropriately placed and nicely formatted crystal
        direction labels [uvw] or [UVTW] to the sector corners.
        """
        fs = self._symmetry.fundamental_sector
        vertices = fs.vertices
        if vertices.size > 0:
            center = fs.center.y[0]

            # Nicely formatted labels for crystal directions
            labels = _get_ipf_axes_labels(vertices, self._symmetry)
            x, y = self._projection.vector2xy(vertices)

            x_edge, y_edge = self._edge_patch.get_path().vertices.T
            x_min_edge, x_max_edge = np.min(x_edge), np.max(x_edge)
            y_min_edge, y_max_edge = np.min(y_edge), np.max(y_edge)
            pad = 0.01
            x_pad = pad * (x_max_edge - x_min_edge)
            y_pad = pad * (y_max_edge - y_min_edge)

            font_size = plt.rcParams["font.size"] + 4
            text_kw = dict(fontsize=font_size, zorder=10)
            for label, xi, yi in zip(labels, x, y):
                # Determine x and y coordinates of label relative to the
                # sector center, and adjust the alignment (starting
                # point) and shift accordingly, so that it doesn't cross
                # over the sector edge
                ha = "center"
                if np.isclose(yi, center, atol=1e-2):
                    va = "center"
                    if np.isclose(xi, x_min_edge, atol=1e-2):
                        ha = "right"
                        xi -= x_pad
                    else:
                        ha = "left"
                        xi += x_pad
                elif yi > center:
                    va = "bottom"
                    # Extra padding ensures [111] in symmetry Th is
                    # placed outside sector edge
                    yi = yi + y_pad + (y_max_edge - yi) * 0.2
                else:
                    va = "top"
                    yi -= y_pad

                maxes.Axes.text(self, xi, yi, s=label, va=va, ha=ha, **text_kw)

    def _pretransform_input_ipf(
        self, values: Union[Tuple[np.ndarray, np.ndarray], Orientation, Vector3d]
    ) -> Vector3d:
        """Return unit vectors within the inverse pole figure from input
        data.

        A call to
        :meth:`orix.plot.StereographicPlot._pretransform_input` after
        this method is required to obtain cartesian coordinates to pass
        to Matplotlib's methods.

        Parameters
        ----------
        values
            Spherical coordinates (azimuth, polar), orientations, or
            vectors. If spherical coordinates are given, they are
            assumed to describe unit vectors. Vectors will be made into
            unit vectors if they aren't already. If orientations are
            passed, the crystal directions returned are the sample
            direction rotated by the orientations.

        Returns
        -------
        v
            Crystal directions.
        """
        if len(values) == 2:  # (Azimuth, polar)
            v = Vector3d.from_polar(azimuth=values[0], polar=values[1])
        elif isinstance(values[0], Vector3d):
            v = values[0]
        else:  # Orientation
            v = values[0] * self._direction
        return v.in_fundamental_sector(self._symmetry)

    def plot_ipf_color_key(self, show_title: bool = True) -> None:
        """Plot an IPF color key code on this axis.

        Parameters
        ----------
        show_title
            Whether to display the Laue group name as the axes title.
            Default is ``True``.

        Notes
        -----
        This function may be used to plot the IPF color key alongside
        another plot where the same key was used to color
        :class:`~orix.quaternion.Orientation` or
        :class:`~orix.vector.Vector3d`.
        """
        symmetry = self._symmetry
        direction_color_key = DirectionColorKeyTSL(symmetry)

        rgba_grid, (x_lim, y_lim) = direction_color_key._create_rgba_grid(
            return_extent=True
        )

        if show_title:
            label_xy = np.column_stack(
                self._projection.vector2xy(symmetry.fundamental_sector.vertices)
            )
            loc = None
            if label_xy.size != 0:
                # Expected title position
                expected_xy = np.array(
                    [np.diff(self.get_xlim())[0] / 2, np.max(self.get_ylim())]
                )
                is_close = np.isclose(label_xy, expected_xy, atol=0.1).all(axis=1)
                if any(is_close) and plt.rcParams["axes.titley"] is None:
                    loc = "left"
            self.set_title(symmetry.name, loc=loc, fontweight="bold")

        self.stereographic_grid(False)
        self._edge_patch.set_linewidth(1.5)
        self.imshow(rgba_grid, extent=x_lim + y_lim, zorder=0)


mprojections.register_projection(InversePoleFigurePlot)


def _setup_inverse_pole_figure_plot(
    symmetry: Symmetry,
    direction: Optional[Vector3d] = None,
    hemisphere: Optional[str] = None,
    figure_kwargs: Optional[Dict] = None,
) -> Tuple[Figure, np.ndarray]:
    """Set up an inverse pole figure plot.

    Parameters
    ----------
    symmetry
        Laue group symmetry of crystal to plot directions with.
    direction
        Sample direction to plot with respect to crystal directions. If
        not given, the out of plane direction, sample Z, is used.
    hemisphere
        Which hemisphere(s) to plot the vectors in. If not given,
        ``"upper"`` is used. Options are ``"upper"``, ``"lower"``, and
        ``"both"``, which plots two projections side by side.
    figure_kwargs
        Dictionary of keyword arguments passed to
        :func:`matplotlib.pyplot.figure`.

    Returns
    -------
    figure
    axes
    """
    if direction is None:
        direction = Vector3d.zvector()

    n_plots = direction.size
    add_hemisphere_label = False
    if hemisphere is None:
        hemisphere = ["upper"] * n_plots
    elif hemisphere == "both":
        add_hemisphere_label = True
        hemisphere = ["upper"] * n_plots + ["lower"] * n_plots
        n_plots *= 2
        direction = Vector3d.stack((direction, direction))
    else:
        # Make iterable
        hemisphere = [hemisphere] * n_plots

    direction = direction.flatten()

    if n_plots <= 3:
        nrows = 1
        ncols = n_plots
    else:
        ncols = 3
        nrows = int(np.ceil(n_plots / 3))

    if figure_kwargs is None:
        figure_kwargs = {"layout": "tight"}
    figure = plt.figure(**figure_kwargs)
    axes = []
    subplot_kw = dict(projection="ipf", symmetry=symmetry)
    for i in range(n_plots):
        subplot_kw.update(dict(direction=direction[i], hemisphere=hemisphere[i]))
        ax = figure.add_subplot(nrows, ncols, i + 1, **subplot_kw)

        label_xy = np.column_stack(
            ax._projection.vector2xy(symmetry.fundamental_sector.vertices)
        )
        loc = None
        if label_xy.size != 0:
            # Expected title position
            expected_xy = np.array(
                [np.diff(ax.get_xlim())[0] / 2, np.max(ax.get_ylim())]
            )
            is_close = np.isclose(label_xy, expected_xy, atol=0.1).all(axis=1)
            if any(is_close) and plt.rcParams["axes.titley"] is None:
                loc = "left"

        ax.set_title(_get_ipf_title(direction[i]), loc=loc, fontweight="bold")

        if add_hemisphere_label:
            ax.show_hemisphere_label()

        axes.append(ax)

    return figure, np.asarray(axes)


def _get_ipf_title(direction: Vector3d) -> str:
    """Get a nicely formatted sample direction string from vector
    coordinates.

    Parameters
    ----------
    direction
        Single vector denoting the sample direction.

    Returns
    -------
    str
    """
    v = Vector3d(((1, 0, 0), (0, 1, 0), (0, 0, 1)))
    idx = np.where(np.isclose(direction.unit.dot(v), 1))[0]
    if idx.size != 0:
        return ["x", "y", "z"][idx[0]]
    else:
        return np.array_str(direction.data.squeeze()).strip("[]")


def _get_ipf_axes_labels(vertices: Vector3d, symmetry: Symmetry) -> List[str]:
    r"""Get nicely formatted crystal direction strings from vector
    coordinates.

    Parameters
    ----------
    vertices
    symmetry
        Symmetry to determine which crystal directions `vertices`
        represent.

    Returns
    -------
    list of str
        List of strings, with -1 formatted like $\bar{1}$.
    """
    phase = Phase(point_group=symmetry)
    m = Miller(uvw=vertices.data, phase=phase)
    if symmetry.system in ["trigonal", "hexagonal"]:
        m.coordinate_format = "UVTW"
    axes = m.round(max_index=2).coordinates.astype(int)

    labels = []
    for ax in axes:
        label = r"[$"
        for i in ax:
            idx = str(abs(i))
            if i < 0:
                label += r"\bar{" + idx + r"}"
            else:
                label += idx
            label += " "
        label = label[:-1] + r"$]"
        labels.append(label)

    return labels



"""
================================================
Rotations mapping the fundamental sector on *S2*
================================================

This example shows how to sample rotations :math:`\mathbf{R}` that when rotating the
vector :math:`\mathbf{v_z} = (0, 0, 1)`, the resulting vectors cover the fundamental
sector of a given Laue class.

We show this by comparing the vectors we get by: 

1. Sampling rotations for *4/mmm* and then rotating :math:`\mathbf{v_z}`
2. Sampling all of *S2* but only keeping those within the corresponding fundamental
   sector.

Apart from the first rotation, all rotations have a Euler angle
:math:`\phi = 0^{\circ}`.
These "reduced" rotations can be useful in template matching of spot patterns from the
transmission electron microscope.
"""

import matplotlib.pyplot as plt
import numpy as np

from orix import plot, sampling
from orix.quaternion import symmetry
from orix.vector import Vector3d

# Sample rotations with an average misorientation
res = 2
pg = symmetry.D4h  # 4/mmm

R = sampling.get_sample_reduced_fundamental(res, point_group=pg)
print(np.allclose(R.to_euler()[1:, 0], 0))

########################################################################
# Get vectors within the fundamental sector following the two routes
v1 = R * Vector3d.zvector()

v2 = sampling.sample_S2(res)
v2 = v2[v2 <= pg.fundamental_sector]

# Only equivalent for the same S2 sampling method
print(np.allclose(v1.data, v2.data))
print(v1)
print(v2)

########################################################################
# Plot the vectors in the fundamental sector of Laue group 4/mmm
fig, (ax0, ax1) = plt.subplots(
    ncols=2, subplot_kw={"projection": "ipf", "symmetry": pg}, layout="tight"
)
ax0.scatter(v1, s=5)
ax1.scatter(v2, c="C1", s=5)
ax0.set_title("Rotated Z vectors", loc="left")
ax1.set_title("Directly sampled", loc="left")
_ = fig.suptitle("Vectors in the fundamental sector of 4/mmm", y=0.8)



# -*- coding: utf-8 -*-
# Copyright 2018-2024 the orix developers
#
# This file is part of orix.
#
# orix is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# orix is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with orix.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import annotations

from itertools import product as iproduct
from typing import Any, List, Optional, Tuple, Union
import warnings

import dask.array as da
from dask.diagnostics import ProgressBar
from matplotlib.gridspec import SubplotSpec
import matplotlib.pyplot as plt
import numpy as np
from scipy.spatial.transform import Rotation as SciPyRotation
from tqdm import tqdm

from orix.quaternion.orientation_region import OrientationRegion
from orix.quaternion.rotation import Rotation
from orix.quaternion.symmetry import C1, Symmetry, _get_unique_symmetry_elements
from orix.vector import Miller


class Misorientation(Rotation):
    r"""Misorientations :math:`M`.

    Misorientations represent transformations from one orientation,
    :math:`O_1` to another, :math:`O_2`: :math:`O_2 \cdot O_1^{-1}`.

    They have symmetries associated with each of the starting
    orientations.

    Parameters
    ----------
    data
        Quaternions.
    symmetry
        Crystal symmetries.
    """

    _symmetry = (C1, C1)

    def __init__(
        self,
        data: Union[np.ndarray, Misorientation, list, tuple],
        symmetry: Optional[Tuple[Symmetry, Symmetry]] = None,
    ):
        super().__init__(data)
        if symmetry:
            self.symmetry = symmetry

    # -------------------------- Properties -------------------------- #

    @property
    def symmetry(self) -> Tuple[Symmetry, Symmetry]:
        """Return or set the crystal symmetries.

        Parameters
        ----------
        value : list of Symmetry or 2-tuple of Symmetry
            Crystal symmetries.
        """
        return self._symmetry

    @symmetry.setter
    def symmetry(self, value: Union[List[Symmetry], Tuple[Symmetry, Symmetry]]):
        if not isinstance(value, (list, tuple)):
            raise TypeError("Value must be a 2-tuple of Symmetry objects.")
        if len(value) != 2 or not all(isinstance(s, Symmetry) for s in value):
            raise ValueError("Value must be a 2-tuple of Symmetry objects.")
        self._symmetry = tuple(value)

    # ------------------------ Dunder methods ------------------------ #

    def __eq__(self, other: Union[Any, Misorientation]) -> bool:
        v1 = super().__eq__(other)
        if not v1:
            return v1
        else:
            # Check whether symmetries also are equivalent
            v2 = []
            for sym_s, sym_o in zip(self._symmetry, other._symmetry):
                v2.append(sym_s == sym_o)
            return all(v2)

    def __getitem__(self, key) -> Misorientation:
        M = super().__getitem__(key)
        M._symmetry = self._symmetry
        return M

    def __invert__(self) -> Misorientation:
        M = super().__invert__()
        M._symmetry = self._symmetry[::-1]
        return M

    def __repr__(self):
        """String representation."""
        cls = self.__class__.__name__
        shape = str(self.shape)
        s1, s2 = self._symmetry[0].name, self._symmetry[1].name
        s2 = "" if s2 == "1" else s2
        symm = s1 + (s2 and ", ") + s2
        data = np.array_str(self.data, precision=4, suppress_small=True)
        rep = "{} {} {}\n{}".format(cls, shape, symm, data)
        return rep

    # ------------------------ Class methods ------------------------- #

    @classmethod
    def from_align_vectors(
        cls,
        other: Miller,
        initial: Miller,
        weights: Optional[np.ndarray] = None,
        return_rmsd: bool = False,
        return_sensitivity: bool = False,
    ) -> Union[
        Misorientation,
        Tuple[Misorientation, float],
        Tuple[Misorientation, np.ndarray],
        Tuple[Misorientation, float, np.ndarray],
    ]:
        """Return an estimated misorientation to optimally align two
        sets of vectors, one set in each crystal.

        This method wraps
        :meth:`~scipy.spatial.transform.Rotation.align_vectors`. See
        that method for further explanations of parameters and returns.

        Parameters
        ----------
        other
            Directions of shape ``(n,)`` in the other crystal.
        initial
            Directions of shape ``(n,)`` in the initial crystal.
        weights
            Relative importance of the different vectors.
        return_rmsd
            Whether to return the (weighted) root mean square distance
            between ``other`` and ``initial`` after alignment. Default
            is ``False``.
        return_sensitivity
            Whether to return the sensitivity matrix. Default is
            ``False``.

        Returns
        -------
        estimated_misorientation
            Best estimate of the misorientation that transforms
            ``initial`` to ``other``. The symmetry of the misorientation
            is inferred from the phase of ``other`` and ``initial``, if
            given.
        rmsd
            Returned when ``return_rmsd=True``.
        sensitivity
            Returned when ``return_sensitivity=True``.

        Raises
        ------
        ValueError
            If ``other`` and ``initial`` are not Miller instances.

        Examples
        --------
        >>> from orix.quaternion import Misorientation
        >>> from orix.vector import Miller
        >>> from orix.crystal_map import Phase
        >>> t1 = Miller(uvw=[[1, 0, 0], [0, 1, 0]], phase=Phase(point_group="m-3m"))
        >>> t2 = Miller(uvw=[[1, 0, 0], [0, 0, 1]], phase=Phase(point_group="m-3m"))
        >>> M12 = Misorientation.from_align_vectors(t2, t1)
        >>> M12 * t1
        Miller (2,), point group m-3m, uvw
        [[1. 0. 0.]
         [0. 0. 1.]]
        """
        if not isinstance(other, Miller) or not isinstance(initial, Miller):
            raise ValueError(
                "Arguments other and initial must both be of type Miller, "
                f"but are of type {type(other)} and {type(initial)}."
            )

        out = super().from_align_vectors(
            other=other,
            initial=initial,
            weights=weights,
            return_rmsd=return_rmsd,
            return_sensitivity=return_sensitivity,
        )
        out = list(out)

        try:
            out[0].symmetry = (initial.phase.point_group, other.phase.point_group)
        except (AttributeError, ValueError):
            pass

        return out[0] if len(out) == 1 else tuple(out)

    @classmethod
    def from_scipy_rotation(
        cls,
        rotation: SciPyRotation,
        symmetry: Optional[Tuple[Symmetry, Symmetry]] = None,
    ) -> Misorientation:
        """Return misorientationss from
        :class:`scipy.spatial.transform.Rotation`.

        Parameters
        ----------
        rotation
            SciPy rotations.
        symmetry
            Tuple of two sets of crystal symmetries. If not given, the
            returned misorientations are assumed to be transformations
            between crystals with only the identity operation, *1*
            (*C1*).

        Returns
        -------
        M
            Misorientations.

        Notes
        -----
        The SciPy rotations are inverted to be consistent with the orix
        framework of rotations.

        Examples
        --------
        >>> from orix.crystal_map import Phase
        >>> from orix.quaternion import Misorientation, symmetry
        >>> from orix.vector import Miller
        >>> from scipy.spatial.transform import Rotation as SciPyRotation
        >>> R_scipy = SciPyRotation.from_euler("ZXZ", [90, 0, 0], degrees=True)
        >>> M = Misorientation.from_scipy_rotation(
        ...     R_scipy, (symmetry.Oh, symmetry.Oh)
        ... )
        >>> t = Miller(uvw=[1, 1, 0], phase=Phase(point_group="m-3m"))
        >>> R_scipy.apply(t.data)
        array([[-1.,  1.,  0.]])
        >>> M * t
        Miller (1,), point group m-3m, uvw
        [[ 1. -1.  0.]]
        >>> ~M * t
        Miller (1,), point group m-3m, uvw
        [[-1.  1.  0.]]
        """
        M = super().from_scipy_rotation(rotation)
        if symmetry:
            M.symmetry = symmetry
        return M

    @classmethod
    def random(
        cls,
        shape: Union[int, tuple] = 1,
        symmetry: Optional[Tuple[Symmetry, Symmetry]] = None,
    ) -> Misorientation:
        """Create random misorientations.

        Parameters
        ----------
        shape
            Shape of the misorientations.
        symmetry
            Tuple of two sets of crystal symmetries. If not given, the
            returned misorientation(s) is assumed to be transformation
            between crystals with only the identity operation, *1*
            (*C1*).

        Returns
        -------
        M
            Random misorientations.
        """
        M = super().random(shape)
        if symmetry:
            M.symmetry = symmetry
        return M

    # --------------------- Other public methods --------------------- #

    def reshape(self, *shape) -> Misorientation:
        M = super().reshape(*shape)
        M._symmetry = self._symmetry
        return M

    def flatten(self) -> Misorientation:
        M = super().flatten()
        M._symmetry = self._symmetry
        return M

    def squeeze(self) -> Misorientation:
        M = super().squeeze()
        M._symmetry = self._symmetry
        return M

    def transpose(self, *axes) -> Misorientation:
        M = super().transpose(*axes)
        M._symmetry = self._symmetry
        return M

    def equivalent(self, grain_exchange: bool = False) -> Misorientation:
        r"""Return the equivalent misorientations.

        Parameters
        ----------
        grain_exchange
            If ``True`` the rotation :math:`g` and :math:`g^{-1}` are
            considered to be identical. Default is ``False``.

        Returns
        -------
        M
            The equivalent misorientations.
        """
        Gl, Gr = self._symmetry

        if grain_exchange and (Gl._tuples == Gr._tuples):
            M = Misorientation.stack((self, ~self)).flatten()
        else:
            M = Misorientation(self)

        equivalent = Gr.outer(M.outer(Gl))

        return self.__class__(equivalent).flatten()

    def map_into_symmetry_reduced_zone(self, verbose: bool = False) -> Misorientation:
        """Return equivalent transformations which have the smallest
        angle of rotation as a new misorientation.

        Parameters
        ----------
        verbose
            Whether to print a progressbar. Default is ``False``.

        Returns
        -------
        M
            A new misorientation object with the assigned symmetry.

        Examples
        --------
        >>> from orix.quaternion.symmetry import C4, C2
        >>> data = np.array([[0.5, 0.5, 0.5, 0.5], [0, 1, 0, 0]])
        >>> M = Misorientation(data)
        >>> M.symmetry = (C4, C2)
        >>> M.map_into_symmetry_reduced_zone()
        Misorientation (2,) 4, 2
        [[-0.7071  0.7071  0.      0.    ]
        [ 0.      1.      0.      0.    ]]
        """
        Gl, Gr = self._symmetry
        symmetry_pairs = iproduct(Gl, Gr)
        if verbose:
            symmetry_pairs = tqdm(symmetry_pairs, total=Gl.size * Gr.size)

        orientation_region = OrientationRegion.from_symmetry(Gl, Gr)
        o_inside = self.__class__.identity(self.shape)
        outside = np.ones(self.shape, dtype=bool)
        for gl, gr in symmetry_pairs:
            o_transformed = gl * self[outside] * gr
            o_inside[outside] = o_transformed
            outside = ~(o_inside < orientation_region)
            if not np.any(outside):
                break
        o_inside._symmetry = (Gl, Gr)
        return o_inside

    def scatter(
        self,
        projection: str = "axangle",
        figure: Optional[plt.Figure] = None,
        position: Union[int, Tuple[int, int], SubplotSpec] = (1, 1, 1),
        return_figure: bool = False,
        wireframe_kwargs: Optional[dict] = None,
        size: Optional[int] = None,
        figure_kwargs: Optional[dict] = None,
        **kwargs,
    ) -> plt.Figure:
        """Plot misorientations in axis-angle space or the Rodrigues
        fundamental zone.

        Parameters
        ----------
        projection
            Which misorientation space to plot misorientations in,
            either ``"axangle"`` (default) or ``"rodrigues"``.
        figure
            If given, a new plot axis :class:`~orix.plot.AxAnglePlot` or
            :class:`~orix.plot.RodriguesPlot` is added to the figure in
            the position specified by ``position``. If not given, a new
            figure is created.
        position
            Where to add the new plot axis. 121 or (1, 2, 1) places it
            in the first of two positions in a grid of 1 row and 2
            columns. See :meth:`~matplotlib.figure.Figure.add_subplot`
            for further details. Default is (1, 1, 1).
        return_figure
            Whether to return the figure. Default is ``False``.
        wireframe_kwargs
            Keyword arguments passed to
            :meth:`orix.plot.AxAnglePlot.plot_wireframe` or
            :meth:`orix.plot.RodriguesPlot.plot_wireframe`.
        size
            If not given, all misorientations are plotted. If given, a
            random sample of this ``size`` of the misorientations is
            plotted.
        figure_kwargs
            Dictionary of keyword arguments passed to
            :func:`matplotlib.pyplot.figure` if ``figure`` is not given.
        **kwargs
            Keyword arguments passed to
            :meth:`orix.plot.AxAnglePlot.scatter` or
            :meth:`orix.plot.RodriguesPlot.scatter`.

        Returns
        -------
        figure
            Figure with the added plot axis, if ``return_figure=True``.

        See Also
        --------
        orix.plot.AxAnglePlot
        orix.plot.RodriguesPlot
        """
        from orix.plot.rotation_plot import _setup_rotation_plot

        figure, ax = _setup_rotation_plot(
            figure=figure,
            projection=projection,
            position=position,
            figure_kwargs=figure_kwargs,
        )

        # Plot wireframe
        if wireframe_kwargs is None:
            wireframe_kwargs = {}
        if isinstance(self.symmetry, tuple):
            fundamental_zone = OrientationRegion.from_symmetry(
                s1=self.symmetry[0], s2=self.symmetry[1]
            )
            ax.plot_wireframe(fundamental_zone, **wireframe_kwargs)
        else:
            # Orientation via inheritance
            fundamental_zone = OrientationRegion.from_symmetry(self.symmetry)
            ax.plot_wireframe(fundamental_zone, **wireframe_kwargs)

        # Correct the aspect ratio of the axes according to the extent
        # of the boundaries of the fundamental region, and also restrict
        # the data limits to these boundaries
        ax._correct_aspect_ratio(fundamental_zone)

        ax.axis("off")
        figure.subplots_adjust(left=0, right=1, bottom=0, top=1, hspace=0, wspace=0)

        if size is not None:
            to_plot = self.get_random_sample(size)
        else:
            to_plot = self
        ax.scatter(to_plot, fundamental_zone=fundamental_zone, **kwargs)

        if return_figure:
            return figure

    def get_distance_matrix(
        self,
        chunk_size: int = 20,
        progressbar: bool = True,
        degrees: bool = False,
    ) -> np.ndarray:
        r"""Return the symmetry reduced smallest angle of rotation
        transforming every misorientation in this instance to every
        other misorientation :cite:`johnstone2020density`.

        Parameters
        ----------
        chunk_size
            Number of misorientations per axis to include in each
            iteration of the computation. Default is 20. Increasing this
            might reduce the computation time at the cost of increased
            memory use.
        progressbar
            Whether to show a progressbar during computation. Default is
            ``True``.
        degrees
            If ``True``, the angles are returned in degrees. Default is
            ``False``.

        Returns
        -------
        angles
            Misorientation angles in radians (``degrees=False``) or
            degrees (``degrees=True``).

        Notes
        -----
        Given two misorientations :math:`M_i` and :math:`M_j` with the
        same two symmetry groups, the smallest angle is considered as
        the geodesic distance

        .. math::

            d(M_i, M_j) = \arccos(2(M_i \cdot M_j)^2 - 1),

        where :math:`(M_i \cdot M_j)` is the highest dot product
        between symmetrically equivalent misorientations to
        :math:`M_{i,j}`, given by

        .. math::

            \max_{s_k \in S_k} s_k M_i s_l s_k M_j^{-1} s_l,

        where :math:`s_k \in S_k` and :math:`s_l \in S_l`, with
        :math:`S_k` and :math:`S_l` being the two symmetry groups.

        Examples
        --------
        >>> from orix.quaternion import Misorientation, symmetry
        >>> M = Misorientation.from_axes_angles([1, 0, 0], [0, 90], degrees=True)
        >>> M.symmetry = (symmetry.D6, symmetry.D6)
        >>> M.get_distance_matrix(progressbar=False, degrees=True)
        array([[ 0., 90.],
               [90.,  0.]])
        """
        # Reduce symmetry operations to the unique ones
        symmetry = _get_unique_symmetry_elements(*self.symmetry)

        # Perform "s_k m_i s_l s_k m_j" (see Notes)
        M1 = symmetry.outer(self).outer(symmetry)
        M2 = M1._outer_dask(~self, chunk_size=chunk_size)

        # Perform last outer product and reduce to all dot products at
        # the same time
        warnings.filterwarnings("ignore", category=da.PerformanceWarning)
        str1 = "abcdefghijklmnopqrstuvwxy"[: M2.ndim]
        str2 = "z" + str1[-1]  # Last axis has shape (4,)
        sum_over = f"{str1},{str2}->{str1[:-1] + str2[0]}"
        all_dot_products = da.einsum(sum_over, M2, symmetry.data)

        # Get highest dot product
        axes = (0, self.ndim + 1, 2 * self.ndim + 2)
        dot_products = da.max(abs(all_dot_products), axis=axes)

        # Round because some dot products are slightly above 1
        dot_products = da.round(dot_products, 12)

        # Calculate disorientation angles
        angles_dask = da.arccos(2 * dot_products**2 - 1)
        angles_dask = da.nan_to_num(angles_dask)
        angles = np.zeros(angles_dask.shape)
        if progressbar:
            with ProgressBar():
                da.store(sources=angles_dask, targets=angles)
        else:
            da.store(sources=angles_dask, targets=angles)

        if degrees:
            angles = np.rad2deg(angles)

        return angles

    def inv(self) -> Misorientation:
        r"""Return the inverse misorientations :math:`M^{-1}`."""
        return self.__invert__()
